import threading
import unittest

from gi.repository import GObject
from .helper import capture_glib_warnings

try:
    from gi.repository import Gtk
except ImportError:
    Gtk = None


_lock = threading.Lock()
counter = 0


def get_id():
    global counter
    with _lock:
        counter += 1
        return counter


class EnumTests(unittest.TestCase):
    def test_gtype(self):
        class MyEnum(GObject.GEnum):
            ONE = 1
            TWO = 2
            THREE = 3

        # The new enum has it's own GType, which subclasses GEnum
        self.assertIsInstance(MyEnum.__gtype__, GObject.GType)
        self.assertNotEqual(MyEnum.__gtype__, GObject.GEnum.__gtype__)
        self.assertEqual(MyEnum.__gtype__.parent, GObject.GEnum.__gtype__)
        self.assertTrue(MyEnum.__gtype__.is_a(GObject.GEnum.__gtype__))
        self.assertIn(MyEnum.__gtype__, GObject.GEnum.__gtype__.children)

        # The class can be looked up by name
        type_name = MyEnum.__gtype__.name
        self.assertIn("MyEnum", type_name)
        self.assertEqual(GObject.GType.from_name(type_name), MyEnum.__gtype__)

        # The Python class is registered as the wrapper for the GType
        self.assertIs(MyEnum.__gtype__.pytype, MyEnum)

    def test_values(self):
        class MyEnum(GObject.GEnum):
            ONE = 1
            FORTY_TWO = 42

        # As this is a stdlib enum, the enum values are subclasses of the enum
        self.assertIsInstance(MyEnum.ONE, MyEnum)
        self.assertIsInstance(MyEnum.FORTY_TWO, MyEnum)

        # We can see the registered enum values too:
        self.assertEqual(MyEnum.ONE.value_name, "ONE")
        self.assertEqual(MyEnum.ONE.value_nick, "one")
        self.assertEqual(MyEnum.FORTY_TWO.value_name, "FORTY_TWO")
        self.assertEqual(MyEnum.FORTY_TWO.value_nick, "forty-two")

    def test_custom_type_name(self):
        type_name = f"MyEnum{get_id()}"

        class MyEnum(GObject.GEnum):
            __gtype_name__ = type_name
            ONE = 1

        self.assertEqual(MyEnum.__gtype__.name, type_name)

        # Trying to register a type with the same name fails:
        with (
            self.assertRaises(RuntimeError) as ex,
            capture_glib_warnings(allow_criticals=True) as w,
        ):

            class MyEnum2(GObject.GEnum):
                __gtype_name__ = type_name
                ONE = 1

        self.assertEqual(str(ex.exception), f"Unable to register enum '{type_name}'")
        self.assertEqual(len(w), 1)
        self.assertEqual(
            str(w[0].message), f"cannot register existing type '{type_name}'"
        )

    @unittest.skipUnless(Gtk, "Gtk not available")
    def test_enum_as_string(self):
        box = Gtk.Box(orientation="vertical")

        assert box.props.orientation == Gtk.Orientation.VERTICAL


class FlagsTests(unittest.TestCase):
    def test_gtype(self):
        class MyFlags(GObject.GFlags):
            ONE = 1
            TWO = 2
            FOUR = 4

        # The new enum has it's own GType, which subclasses GFlags
        self.assertIsInstance(MyFlags.__gtype__, GObject.GType)
        self.assertNotEqual(MyFlags.__gtype__, GObject.GFlags.__gtype__)
        self.assertEqual(MyFlags.__gtype__.parent, GObject.GFlags.__gtype__)
        self.assertTrue(MyFlags.__gtype__.is_a(GObject.GFlags.__gtype__))
        self.assertIn(MyFlags.__gtype__, GObject.GFlags.__gtype__.children)

        # The class can be looked up by name
        type_name = MyFlags.__gtype__.name
        self.assertIn("MyFlags", type_name)
        self.assertEqual(GObject.GType.from_name(type_name), MyFlags.__gtype__)

        # The Python class is registered as the wrapper for the GType
        self.assertIs(MyFlags.__gtype__.pytype, MyFlags)

    def test_values(self):
        class MyFlags(GObject.GFlags):
            ONE = 1
            THIRTY_TWO = 32

        # As this is a stdlib enum, the enum values are subclasses of the enum
        self.assertIsInstance(MyFlags.ONE, MyFlags)
        self.assertIsInstance(MyFlags.THIRTY_TWO, MyFlags)

        # We can see the registered enum values too:
        self.assertEqual(MyFlags.ONE.value_names, ["ONE"])
        self.assertEqual(MyFlags.ONE.value_nicks, ["one"])
        self.assertEqual(MyFlags.THIRTY_TWO.value_names, ["THIRTY_TWO"])
        self.assertEqual(MyFlags.THIRTY_TWO.value_nicks, ["thirty-two"])

        # Similar for combinations of flags
        v = MyFlags.ONE | MyFlags.THIRTY_TWO
        self.assertIsInstance(v, MyFlags)
        self.assertEqual(v.value_names, ["ONE", "THIRTY_TWO"])
        self.assertEqual(v.value_nicks, ["one", "thirty-two"])

    def test_custom_type_name(self):
        type_name = f"MyFlags{get_id()}"

        class MyFlags(GObject.GFlags):
            __gtype_name__ = type_name
            ONE = 1

        self.assertEqual(MyFlags.__gtype__.name, type_name)

        # Trying to register a type with the same name fails:
        with (
            self.assertRaises(RuntimeError) as ex,
            capture_glib_warnings(allow_criticals=True) as w,
        ):

            class MyFlags2(GObject.GFlags):
                __gtype_name__ = type_name
                ONE = 1

        self.assertEqual(str(ex.exception), f"Unable to register flags '{type_name}'")
        self.assertEqual(len(w), 1)
        self.assertEqual(
            str(w[0].message), f"cannot register existing type '{type_name}'"
        )



import os
import gc
import unittest
import tempfile
import socket

import pytest

import gi
from gi.repository import GLib

from .helper import capture_gi_deprecation_warnings


def test_io_add_watch_get_args():
    get_args = GLib._io_add_watch_get_args

    def func():
        return None

    # create a closed channel for testing
    fd, fn = tempfile.mkstemp()
    os.close(fd)
    try:
        chan = GLib.IOChannel(filename=fn)
        chan.shutdown(True)
    finally:
        os.remove(fn)

    # old way
    with capture_gi_deprecation_warnings():
        assert get_args(chan, GLib.IOCondition.IN, func) == (
            chan,
            0,
            GLib.IOCondition.IN,
            func,
            (),
        )

        with pytest.raises(TypeError):
            get_args(chan, GLib.IOCondition.IN, object())

    # new way
    prio = GLib.PRIORITY_DEFAULT
    with capture_gi_deprecation_warnings():
        assert get_args(chan, prio, GLib.IOCondition.IN, func, 99) == (
            chan,
            prio,
            GLib.IOCondition.IN,
            func,
            (99,),
        )

        with pytest.raises(TypeError):
            assert get_args(chan, prio, GLib.IOCondition.IN)

        with pytest.raises(TypeError):
            assert get_args(chan, prio, 99)


@pytest.mark.skipif(os.name != "nt", reason="windows only")
def test_io_add_watch_get_args_win32_socket():
    get_args = GLib._io_add_watch_get_args

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def func():
        return None

    prio = GLib.PRIORITY_DEFAULT
    chan = get_args(s, prio, GLib.IOCondition.IN, func)[0]
    assert isinstance(chan, GLib.IOChannel)
    chan.shutdown(False)


def test_threads_init():
    with capture_gi_deprecation_warnings() as w:
        GLib.threads_init()
    assert len(w)


def test_gerror_matches():
    e = GLib.Error(domain=42, code=24)
    assert e.matches(42, 24)


def test_timeout_add_seconds():
    h = GLib.timeout_add_seconds(
        100, lambda *x: None, 1, 2, 3, priority=GLib.PRIORITY_HIGH_IDLE
    )
    GLib.source_remove(h)


def test_iochannel():
    with pytest.raises(TypeError):
        GLib.IOChannel()


def test_iochannel_write():
    fd, fn = tempfile.mkstemp()
    os.close(fd)
    chan = GLib.IOChannel(filename=fn, mode="r+")
    try:
        assert chan.write(b"foo", 2) == 2
        chan.seek(0)
        assert chan.read() == b"fo"
    finally:
        chan.shutdown(True)


@pytest.mark.skipif(os.name == "nt", reason="unix only")
def test_has_unix_signal_add():
    with capture_gi_deprecation_warnings():
        GLib.unix_signal_add == GLib.unix_signal_add_full


@pytest.mark.skipif(os.name != "nt", reason="windows only")
def test_iochannel_win32():
    fd, fn = tempfile.mkstemp()
    closed = False
    try:
        channel = GLib.IOChannel(hwnd=fd)
        try:
            assert channel.read() == b""
        finally:
            closed = True
            channel.shutdown(True)
    finally:
        if not closed:
            os.close(fd)
        os.remove(fn)


class TestGVariant(unittest.TestCase):
    def test_create_simple(self):
        variant = GLib.Variant("i", 42)
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.get_int32(), 42)

        variant = GLib.Variant("s", "")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.get_string(), "")

        variant = GLib.Variant("s", "hello")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.get_string(), "hello")

    def test_simple_invalid_ops(self):
        variant = GLib.Variant("i", 42)
        with pytest.raises(TypeError):
            len(variant)

        with pytest.raises(TypeError):
            variant[0]

        with pytest.raises(TypeError):
            variant.keys()

    def test_create_variant(self):
        variant = GLib.Variant("v", GLib.Variant("i", 42))
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertTrue(isinstance(variant.get_variant(), GLib.Variant))
        self.assertEqual(variant.get_type_string(), "v")
        self.assertEqual(variant.get_variant().get_type_string(), "i")
        self.assertEqual(variant.get_variant().get_int32(), 42)

        variant = GLib.Variant("v", GLib.Variant("v", GLib.Variant("i", 42)))
        self.assertEqual(variant.get_type_string(), "v")
        self.assertEqual(variant.get_variant().get_type_string(), "v")
        self.assertEqual(variant.get_variant().get_variant().get_type_string(), "i")
        self.assertEqual(variant.get_variant().get_variant().get_int32(), 42)

    def test_create_tuple(self):
        variant = GLib.Variant("()", ())
        self.assertEqual(variant.get_type_string(), "()")
        self.assertEqual(variant.n_children(), 0)

        variant = GLib.Variant("(i)", (3,))
        self.assertEqual(variant.get_type_string(), "(i)")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.n_children(), 1)
        self.assertTrue(isinstance(variant.get_child_value(0), GLib.Variant))
        self.assertEqual(variant.get_child_value(0).get_int32(), 3)

        variant = GLib.Variant("(ss)", ("mec", "mac"))
        self.assertEqual(variant.get_type_string(), "(ss)")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(0), GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(1), GLib.Variant))
        self.assertEqual(variant.get_child_value(0).get_string(), "mec")
        self.assertEqual(variant.get_child_value(1).get_string(), "mac")

        # nested tuples
        variant = GLib.Variant("((si)(ub))", (("hello", -1), (42, True)))
        self.assertEqual(variant.get_type_string(), "((si)(ub))")
        self.assertEqual(variant.unpack(), (("hello", -1), (42, True)))

    def test_new_tuple_sink(self):
        # https://bugzilla.gnome.org/show_bug.cgi?id=735166
        variant = GLib.Variant.new_tuple(GLib.Variant.new_tuple())
        del variant
        gc.collect()

    def test_create_dictionary(self):
        variant = GLib.Variant("a{si}", {})
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.get_type_string(), "a{si}")
        self.assertEqual(variant.n_children(), 0)

        variant = GLib.Variant("a{si}", {"": 1, "key1": 2, "key2": 3})
        self.assertEqual(variant.get_type_string(), "a{si}")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(0), GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(1), GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(2), GLib.Variant))
        self.assertEqual(variant.unpack(), {"": 1, "key1": 2, "key2": 3})

        # nested dictionaries
        variant = GLib.Variant("a{sa{si}}", {})
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.get_type_string(), "a{sa{si}}")
        self.assertEqual(variant.n_children(), 0)

        d = {"": {"": 1, "keyn1": 2}, "key1": {"key11": 11, "key12": 12}}
        variant = GLib.Variant("a{sa{si}}", d)
        self.assertEqual(variant.get_type_string(), "a{sa{si}}")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertEqual(variant.unpack(), d)

        # init with an iterable
        variant = GLib.Variant("a{si}", [("foo", 2)])
        assert variant.unpack() == {"foo": 2}

        with pytest.raises(TypeError):
            GLib.Variant("a{si}", [("foo",)])
        with pytest.raises(TypeError):
            GLib.Variant("a{si}", [("foo", 1, 2)])

    def test_create_array(self):
        variant = GLib.Variant("ai", [])
        self.assertEqual(variant.get_type_string(), "ai")
        self.assertEqual(variant.n_children(), 0)

        variant = GLib.Variant("ai", [1, 2])
        self.assertEqual(variant.get_type_string(), "ai")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(0), GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(1), GLib.Variant))
        self.assertEqual(variant.get_child_value(0).get_int32(), 1)
        self.assertEqual(variant.get_child_value(1).get_int32(), 2)

        variant = GLib.Variant("as", [])
        self.assertEqual(variant.get_type_string(), "as")
        self.assertEqual(variant.n_children(), 0)

        variant = GLib.Variant("as", [""])
        self.assertEqual(variant.get_type_string(), "as")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(0), GLib.Variant))
        self.assertEqual(variant.get_child_value(0).get_string(), "")

        variant = GLib.Variant("as", ["hello", "world"])
        self.assertEqual(variant.get_type_string(), "as")
        self.assertTrue(isinstance(variant, GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(0), GLib.Variant))
        self.assertTrue(isinstance(variant.get_child_value(1), GLib.Variant))
        self.assertEqual(variant.get_child_value(0).get_string(), "hello")
        self.assertEqual(variant.get_child_value(1).get_string(), "world")

        # nested arrays
        variant = GLib.Variant("aai", [])
        self.assertEqual(variant.get_type_string(), "aai")
        self.assertEqual(variant.n_children(), 0)

        variant = GLib.Variant("aai", [[]])
        self.assertEqual(variant.get_type_string(), "aai")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 0)

        variant = GLib.Variant("aai", [[1, 2], [3, 4, 5]])
        self.assertEqual(variant.get_type_string(), "aai")
        self.assertEqual(variant.unpack(), [[1, 2], [3, 4, 5]])

    def test_create_array_guchar(self):
        variant = GLib.Variant("ay", [97, 97, 97])
        assert variant.unpack() == [97, 97, 97]

        variant = GLib.Variant("ay", b"aaa")
        assert variant.unpack() == [97, 97, 97]

        variant = GLib.Variant("ay", iter([1, 2, 3]))
        assert variant.unpack() == [1, 2, 3]

        variant = GLib.Variant("ay", "aaa")
        assert variant.unpack() == [97, 97, 97]

        with self.assertRaises(TypeError):
            GLib.Variant("ay", object())

    def test_create_maybe(self):
        variant = GLib.Variant("mai", None)
        self.assertEqual(variant.get_type_string(), "mai")
        self.assertEqual(variant.n_children(), 0)
        self.assertEqual(variant.unpack(), None)

        variant = GLib.Variant("mai", [])
        self.assertEqual(variant.get_type_string(), "mai")
        self.assertEqual(variant.n_children(), 1)

        variant = GLib.Variant("mami", [None])
        self.assertEqual(variant.get_type_string(), "mami")
        self.assertEqual(variant.n_children(), 1)

        variant = GLib.Variant("mami", [None, 13, None])
        self.assertEqual(variant.get_type_string(), "mami")
        self.assertEqual(variant.n_children(), 1)
        array = variant.get_child_value(0)
        self.assertEqual(array.n_children(), 3)

        element = array.get_child_value(0)
        self.assertEqual(element.n_children(), 0)
        element = array.get_child_value(1)
        self.assertEqual(element.n_children(), 1)
        self.assertEqual(element.get_child_value(0).get_int32(), 13)
        element = array.get_child_value(2)
        self.assertEqual(element.n_children(), 0)

    def test_create_complex(self):
        variant = GLib.Variant("(as)", ([],))
        self.assertEqual(variant.get_type_string(), "(as)")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 0)

        variant = GLib.Variant("(as)", ([""],))
        self.assertEqual(variant.get_type_string(), "(as)")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 1)
        self.assertEqual(variant.get_child_value(0).get_child_value(0).get_string(), "")

        variant = GLib.Variant("(as)", (["hello"],))
        self.assertEqual(variant.get_type_string(), "(as)")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 1)
        self.assertEqual(
            variant.get_child_value(0).get_child_value(0).get_string(), "hello"
        )

        variant = GLib.Variant("a(ii)", [])
        self.assertEqual(variant.get_type_string(), "a(ii)")
        self.assertEqual(variant.n_children(), 0)

        variant = GLib.Variant("a(ii)", [(5, 6)])
        self.assertEqual(variant.get_type_string(), "a(ii)")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 2)
        self.assertEqual(variant.get_child_value(0).get_child_value(0).get_int32(), 5)
        self.assertEqual(variant.get_child_value(0).get_child_value(1).get_int32(), 6)

        variant = GLib.Variant("(a(ii))", ([],))
        self.assertEqual(variant.get_type_string(), "(a(ii))")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 0)

        variant = GLib.Variant("(a(ii))", ([(5, 6)],))
        self.assertEqual(variant.get_type_string(), "(a(ii))")
        self.assertEqual(variant.n_children(), 1)
        self.assertEqual(variant.get_child_value(0).n_children(), 1)
        self.assertEqual(variant.get_child_value(0).get_child_value(0).n_children(), 2)
        self.assertEqual(
            variant.get_child_value(0)
            .get_child_value(0)
            .get_child_value(0)
            .get_int32(),
            5,
        )
        self.assertEqual(
            variant.get_child_value(0)
            .get_child_value(0)
            .get_child_value(1)
            .get_int32(),
            6,
        )

        obj = {"a1": (1, True), "a2": (2, False)}
        variant = GLib.Variant("a{s(ib)}", obj)
        self.assertEqual(variant.get_type_string(), "a{s(ib)}")
        self.assertEqual(variant.unpack(), obj)

        obj = {"a1": (1, GLib.Variant("b", True)), "a2": (2, GLib.Variant("y", 255))}
        variant = GLib.Variant("a{s(iv)}", obj)
        self.assertEqual(variant.get_type_string(), "a{s(iv)}")
        self.assertEqual(variant.unpack(), {"a1": (1, True), "a2": (2, 255)})

        obj = (1, {"a": {"a1": True, "a2": False}, "b": {"b1": False}, "c": {}}, "foo")
        variant = GLib.Variant("(ia{sa{sb}}s)", obj)
        self.assertEqual(variant.get_type_string(), "(ia{sa{sb}}s)")
        self.assertEqual(variant.unpack(), obj)

        obj = {
            "frequency": GLib.Variant("t", 738000000),
            "hierarchy": GLib.Variant("i", 0),
            "bandwidth": GLib.Variant("x", 8),
            "code-rate-hp": GLib.Variant("d", 2.0 / 3.0),
            "constellation": GLib.Variant("s", "QAM16"),
            "guard-interval": GLib.Variant("u", 4),
        }
        variant = GLib.Variant("a{sv}", obj)
        self.assertEqual(variant.get_type_string(), "a{sv}")
        self.assertEqual(
            variant.unpack(),
            {
                "frequency": 738000000,
                "hierarchy": 0,
                "bandwidth": 8,
                "code-rate-hp": 2.0 / 3.0,
                "constellation": "QAM16",
                "guard-interval": 4,
            },
        )

    def test_create_errors(self):
        # excess arguments
        self.assertRaises(TypeError, GLib.Variant, "i", 42, 3)
        self.assertRaises(TypeError, GLib.Variant, "(i)", (42, 3))

        # not enough arguments
        self.assertRaises(TypeError, GLib.Variant, "(ii)", (42,))

        # data type mismatch
        self.assertRaises(TypeError, GLib.Variant, "i", "hello")
        self.assertRaises(TypeError, GLib.Variant, "s", 42)
        self.assertRaises(TypeError, GLib.Variant, "(ss)", "mec", "mac")
        self.assertRaises(TypeError, GLib.Variant, "(s)", "hello")

        # invalid format string
        self.assertRaises(TypeError, GLib.Variant, "Q", 1)

        # invalid types
        self.assertRaises(TypeError, GLib.Variant, "(ii", (42, 3))
        self.assertRaises(TypeError, GLib.Variant, "(ii))", (42, 3))
        self.assertRaises(TypeError, GLib.Variant, "a{si", {})
        self.assertRaises(TypeError, GLib.Variant, "a{si}}", {})
        self.assertRaises(TypeError, GLib.Variant, "a{iii}", {})

    def test_unpack(self):
        # simple values
        res = GLib.Variant.new_int32(-42).unpack()
        self.assertEqual(res, -42)

        res = GLib.Variant.new_uint64(34359738368).unpack()
        self.assertEqual(res, 34359738368)

        res = GLib.Variant.new_boolean(True).unpack()
        self.assertEqual(res, True)

        res = GLib.Variant.new_object_path("/foo/Bar").unpack()
        self.assertEqual(res, "/foo/Bar")

        # variant
        res = GLib.Variant("v", GLib.Variant.new_int32(-42)).unpack()
        self.assertEqual(res, -42)

        GLib.Variant("v", GLib.Variant("v", GLib.Variant("i", 42)))
        self.assertEqual(res, -42)

        # tuple
        res = GLib.Variant.new_tuple(
            GLib.Variant.new_int32(-1), GLib.Variant.new_string("hello")
        ).unpack()
        self.assertEqual(res, (-1, "hello"))

        # array
        vb = GLib.VariantBuilder.new(gi._gi.variant_type_from_string("ai"))
        vb.add_value(GLib.Variant.new_int32(-1))
        vb.add_value(GLib.Variant.new_int32(3))
        res = vb.end().unpack()
        self.assertEqual(res, [-1, 3])

        # dictionary
        res = GLib.Variant("a{si}", {"key1": 1, "key2": 2}).unpack()
        self.assertEqual(res, {"key1": 1, "key2": 2})

        # maybe
        v = GLib.Variant("mi", 1)
        self.assertEqual(v.unpack(), 1)
        v = GLib.Variant("mi", None)
        self.assertEqual(v.unpack(), None)
        v = GLib.Variant("mai", [])
        self.assertEqual(v.unpack(), [])
        v = GLib.Variant("m()", ())
        self.assertEqual(v.unpack(), ())
        v = GLib.Variant("mami", [None, 1, None])
        self.assertEqual(v.unpack(), [None, 1, None])

    def test_iteration(self):
        # array index access
        vb = GLib.VariantBuilder.new(gi._gi.variant_type_from_string("ai"))
        vb.add_value(GLib.Variant.new_int32(-1))
        vb.add_value(GLib.Variant.new_int32(3))
        v = vb.end()

        self.assertEqual(len(v), 2)
        self.assertEqual(v[0], -1)
        self.assertEqual(v[1], 3)
        self.assertEqual(v[-1], 3)
        self.assertEqual(v[-2], -1)
        self.assertRaises(IndexError, v.__getitem__, 2)
        self.assertRaises(IndexError, v.__getitem__, -3)
        self.assertRaises(ValueError, v.__getitem__, "a")

        # array iteration
        self.assertEqual(list(v), [-1, 3])
        self.assertEqual(list(v), [-1, 3])

        # tuple index access
        v = GLib.Variant.new_tuple(
            GLib.Variant.new_int32(-1), GLib.Variant.new_string("hello")
        )
        self.assertEqual(len(v), 2)
        self.assertEqual(v[0], -1)
        self.assertEqual(v[1], "hello")
        self.assertEqual(v[-1], "hello")
        self.assertEqual(v[-2], -1)
        self.assertRaises(IndexError, v.__getitem__, 2)
        self.assertRaises(IndexError, v.__getitem__, -3)
        self.assertRaises(ValueError, v.__getitem__, "a")

        # tuple iteration
        self.assertEqual(list(v), [-1, "hello"])
        self.assertEqual(tuple(v), (-1, "hello"))

        # dictionary index access
        vsi = GLib.Variant("a{si}", {"key1": 1, "key2": 2})
        vis = GLib.Variant("a{is}", {1: "val1", 5: "val2"})

        self.assertEqual(len(vsi), 2)
        self.assertEqual(vsi["key1"], 1)
        self.assertEqual(vsi["key2"], 2)
        self.assertRaises(KeyError, vsi.__getitem__, "unknown")

        self.assertEqual(len(vis), 2)
        self.assertEqual(vis[1], "val1")
        self.assertEqual(vis[5], "val2")
        self.assertRaises(KeyError, vsi.__getitem__, 3)

        # dictionary iteration
        self.assertEqual(set(vsi.keys()), {"key1", "key2"})
        self.assertEqual(set(vis.keys()), {1, 5})

        # string index access
        v = GLib.Variant("s", "hello")
        self.assertEqual(len(v), 5)
        self.assertEqual(v[0], "h")
        self.assertEqual(v[4], "o")
        self.assertEqual(v[-1], "o")
        self.assertEqual(v[-5], "h")
        self.assertRaises(IndexError, v.__getitem__, 5)
        self.assertRaises(IndexError, v.__getitem__, -6)

        # string iteration
        self.assertEqual(list(v), ["h", "e", "l", "l", "o"])

    def test_split_signature(self):
        self.assertEqual(GLib.Variant.split_signature("()"), [])

        self.assertEqual(GLib.Variant.split_signature("s"), ["s"])

        self.assertEqual(GLib.Variant.split_signature("as"), ["as"])

        self.assertEqual(GLib.Variant.split_signature("(s)"), ["s"])

        self.assertEqual(GLib.Variant.split_signature("(iso)"), ["i", "s", "o"])

        self.assertEqual(
            GLib.Variant.split_signature("(s(ss)i(ii))"), ["s", "(ss)", "i", "(ii)"]
        )

        self.assertEqual(GLib.Variant.split_signature("(as)"), ["as"])

        self.assertEqual(
            GLib.Variant.split_signature("(s(ss)iaiaasa(ii))"),
            ["s", "(ss)", "i", "ai", "aas", "a(ii)"],
        )

        self.assertEqual(
            GLib.Variant.split_signature("(a{iv}(ii)((ss)a{s(ss)}))"),
            ["a{iv}", "(ii)", "((ss)a{s(ss)})"],
        )

    def test_hash(self):
        v1 = GLib.Variant("s", "somestring")
        v2 = GLib.Variant("s", "somestring")
        v3 = GLib.Variant("s", "somestring2")

        self.assertTrue(v2 in {v1, v3})
        self.assertTrue(v2 in frozenset([v1, v3]))
        self.assertTrue(v2 in {v1: "1", v3: "2"})

    def test_compare(self):
        # Check if identical GVariant are equal

        def assert_equal(vtype, value):
            self.assertEqual(GLib.Variant(vtype, value), GLib.Variant(vtype, value))

        def assert_not_equal(vtype1, value1, vtype2, value2):
            self.assertNotEqual(
                GLib.Variant(vtype1, value1), GLib.Variant(vtype2, value2)
            )

        numbers = ["y", "n", "q", "i", "u", "x", "t", "h", "d"]
        for num in numbers:
            assert_equal(num, 42)
            assert_not_equal(num, 42, num, 41)
            assert_not_equal(num, 42, "s", "42")

        assert_equal("s", "something")
        assert_not_equal("s", "something", "s", "somethingelse")
        assert_not_equal("s", "something", "i", 1234)

        assert_equal("g", "dustybinqhogx")
        assert_not_equal("g", "dustybinqhogx", "g", "dustybin")
        assert_not_equal("g", "dustybinqhogx", "i", 1234)

        assert_equal("o", "/dev/null")
        assert_not_equal("o", "/dev/null", "o", "/dev/zero")
        assert_not_equal("o", "/dev/null", "i", 1234)

        assert_equal("(s)", ("strtuple",))
        assert_not_equal("(s)", ("strtuple",), "(s)", ("strtuple2",))

        assert_equal("a{si}", {"str": 42})
        assert_not_equal("a{si}", {"str": 42}, "a{si}", {"str": 43})

        assert_equal("v", GLib.Variant("i", 42))
        assert_not_equal("v", GLib.Variant("i", 42), "v", GLib.Variant("i", 43))

        assert GLib.Variant("i", 42) != object()
        assert GLib.Variant("i", 42) != object()

    def test_bool(self):
        # Check if the GVariant bool matches the unpacked Pythonic bool

        def assert_equals_bool(vtype, value):
            self.assertEqual(bool(GLib.Variant(vtype, value)), bool(value))

        # simple values
        assert_equals_bool("b", True)
        assert_equals_bool("b", False)

        numbers = ["y", "n", "q", "i", "u", "x", "t", "h", "d"]
        for number in numbers:
            assert_equals_bool(number, 0)
            assert_equals_bool(number, 1)

        assert_equals_bool("s", "")
        assert_equals_bool("g", "")
        assert_equals_bool("s", "something")
        assert_equals_bool("o", "/dev/null")
        assert_equals_bool("g", "dustybinqhogx")

        # arrays
        assert_equals_bool("ab", [True])
        assert_equals_bool("ab", [False])
        for number in numbers:
            assert_equals_bool("a" + number, [])
            assert_equals_bool("a" + number, [0])
        assert_equals_bool("as", [])
        assert_equals_bool("as", [""])
        assert_equals_bool("ao", [])
        assert_equals_bool("ao", ["/"])
        assert_equals_bool("ag", [])
        assert_equals_bool("ag", [""])
        assert_equals_bool("aai", [[]])

        # tuples
        assert_equals_bool("()", ())
        for number in numbers:
            assert_equals_bool("(" + number + ")", (0,))
        assert_equals_bool("(s)", ("",))
        assert_equals_bool("(o)", ("/",))
        assert_equals_bool("(g)", ("",))
        assert_equals_bool("(())", ((),))

        # dictionaries
        assert_equals_bool("a{si}", {})
        assert_equals_bool("a{si}", {"": 0})

        # complex types, always True
        assert_equals_bool("(as)", ([],))
        assert_equals_bool("a{s(i)}", {"": (0,)})

        # variant types, recursive unpacking
        assert_equals_bool("v", GLib.Variant("i", 0))
        assert_equals_bool("v", GLib.Variant("i", 1))

        # maybe types
        assert_equals_bool("mi", 42)
        assert_equals_bool("mi", 0)
        assert_equals_bool("mi", None)

    def test_repr(self):
        # with C constructor
        v = GLib.Variant.new_uint32(42)
        self.assertEqual(repr(v), "GLib.Variant('u', 42)")

        # with override constructor
        v = GLib.Variant("(is)", (1, "somestring"))
        self.assertEqual(repr(v), "GLib.Variant('(is)', (1, 'somestring'))")

    def test_str(self):
        # with C constructor
        v = GLib.Variant.new_uint32(42)
        self.assertEqual(str(v), "uint32 42")

        # with override constructor
        v = GLib.Variant("(is)", (1, "somestring"))
        self.assertEqual(str(v), "(1, 'somestring')")

    def test_parse_error(self):
        # This test doubles as a test for GLib.Error marshaling.
        source_str = "abc"
        with self.assertRaises(GLib.Error) as context:
            GLib.Variant.parse(None, source_str, None, None)
        e = context.exception
        text = GLib.Variant.parse_error_print_context(e, source_str)
        self.assertTrue(source_str in text)

    def test_parse_error_exceptions(self):
        source_str = "abc"
        self.assertRaisesRegex(
            TypeError,
            "Must be GLib.Error, not int",
            GLib.Variant.parse_error_print_context,
            42,
            source_str,
        )

        gerror = GLib.Error(message=42)  # not a string
        self.assertRaisesRegex(
            TypeError,
            ".*Must be string, not int.*",
            GLib.Variant.parse_error_print_context,
            gerror,
            source_str,
        )

        gerror = GLib.Error(domain=42)  # not a string
        self.assertRaisesRegex(
            TypeError,
            ".*Must be string, not int.*",
            GLib.Variant.parse_error_print_context,
            gerror,
            source_str,
        )

        gerror = GLib.Error(code="not an int")
        self.assertRaisesRegex(
            TypeError,
            ".*Must be number, not str.*",
            GLib.Variant.parse_error_print_context,
            gerror,
            source_str,
        )

        gerror = GLib.Error(code=GLib.MAXUINT)
        self.assertRaisesRegex(
            OverflowError,
            ".*not in range.*",
            GLib.Variant.parse_error_print_context,
            gerror,
            source_str,
        )


class TestConstants(unittest.TestCase):
    def test_basic_types_limits(self):
        self.assertTrue(isinstance(GLib.MINFLOAT, float))
        self.assertTrue(isinstance(GLib.MAXLONG, int))



import re
import os
import sys
import unittest
import pytest

import gi
import gi.overrides
from gi import PyGIDeprecationWarning

try:
    from gi.repository import Gio, Gdk, GdkPixbuf, Gtk

    GDK4 = Gdk._version == "4.0"
except ImportError:
    Gdk = None
    Gtk = None
    GDK4 = False


def gtkver():
    if Gtk is None:
        return (0, 0, 0)
    return (Gtk.get_major_version(), Gtk.get_minor_version(), Gtk.get_micro_version())


try:
    gi.require_foreign("cairo")
    has_cairo = True
except ImportError:
    has_cairo = False

from .helper import capture_glib_deprecation_warnings


@unittest.skipUnless(Gdk, "Gdk not available")
class TestGdk(unittest.TestCase):
    @unittest.skipIf(sys.platform == "darwin" or os.name == "nt", "crashes")
    @unittest.skipIf(GDK4, "not in gdk4")
    def test_constructor(self):
        attribute = Gdk.WindowAttr()
        attribute.window_type = Gdk.WindowType.CHILD
        attributes_mask = Gdk.WindowAttributesType.X | Gdk.WindowAttributesType.Y
        window = Gdk.Window(None, attribute, attributes_mask)
        self.assertEqual(window.get_window_type(), Gdk.WindowType.CHILD)

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_color(self):
        color = Gdk.Color(100, 200, 300)
        self.assertEqual(color.red, 100)
        self.assertEqual(color.green, 200)
        self.assertEqual(color.blue, 300)
        with capture_glib_deprecation_warnings():
            self.assertEqual(color, Gdk.Color(100, 200, 300))
        self.assertNotEqual(color, Gdk.Color(1, 2, 3))
        self.assertNotEqual(color, None)
        # assertNotEqual only tests __ne__. Following line explicitly
        # tests __eq__ with objects of other types
        self.assertFalse(color == object())

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_color_floats(self):
        self.assertEqual(
            Gdk.Color(13107, 21845, 65535), Gdk.Color.from_floats(0.2, 1.0 / 3.0, 1.0)
        )

        self.assertEqual(
            Gdk.Color(13107, 21845, 65535).to_floats(), (0.2, 1.0 / 3.0, 1.0)
        )

        self.assertEqual(
            Gdk.RGBA(0.2, 1.0 / 3.0, 1.0, 0.5).to_color(),
            Gdk.Color.from_floats(0.2, 1.0 / 3.0, 1.0),
        )

        self.assertEqual(
            Gdk.RGBA.from_color(Gdk.Color(13107, 21845, 65535)),
            Gdk.RGBA(0.2, 1.0 / 3.0, 1.0, 1.0),
        )

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_color_to_floats_attrs(self):
        color = Gdk.Color(13107, 21845, 65535)
        assert color.red_float == 0.2
        color.red_float = 0
        assert color.red_float == 0
        assert color.green_float == 1.0 / 3.0
        color.green_float = 0
        assert color.green_float == 0
        assert color.blue_float == 1.0
        color.blue_float = 0
        assert color.blue_float == 0

    def test_rgba(self):
        self.assertEqual(Gdk.RGBA, gi.overrides.Gdk.RGBA)
        rgba = Gdk.RGBA(0.1, 0.2, 0.3, 0.4)
        self.assertEqual(rgba, Gdk.RGBA(0.1, 0.2, 0.3, 0.4))
        self.assertNotEqual(rgba, Gdk.RGBA(0.0, 0.2, 0.3, 0.4))
        self.assertAlmostEqual(rgba.red, 0.1)
        self.assertAlmostEqual(rgba.green, 0.2)
        self.assertAlmostEqual(rgba.blue, 0.3)
        self.assertAlmostEqual(rgba.alpha, 0.4)
        rgba.green = 0.9
        self.assertAlmostEqual(rgba.green, 0.9)
        self.assertNotEqual(rgba, None)
        # assertNotEqual only tests __ne__. Following line explicitly
        # tests __eq__ with objects of other types
        self.assertFalse(rgba == object())

        # Iterator/tuple convsersion
        assert tuple(Gdk.RGBA(0.1, 0.2, 0.3, 0.4)) == pytest.approx(
            (0.1, 0.2, 0.3, 0.4)
        )

    @unittest.skipUnless(GDK4, "only in gdk4")
    def test_rgba_gtk4(self):
        c = Gdk.RGBA(0, 0, 0, 0)
        assert c.to_string() == "rgba(0,0,0,0)"

    @unittest.skipIf(not has_cairo or GDK4, "not in gdk4")
    def test_window(self):
        w = Gtk.Window()
        w.realize()
        window = w.get_window()
        with capture_glib_deprecation_warnings():
            assert window.cairo_create() is not None

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_drag_context(self):
        context = Gdk.DragContext()
        # using it this way crashes..
        assert hasattr(context, "finish")

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_event(self):
        event = Gdk.Event.new(Gdk.EventType.CONFIGURE)
        self.assertEqual(event.type, Gdk.EventType.CONFIGURE)
        self.assertEqual(event.send_event, 0)

        event = Gdk.Event()
        event.type = Gdk.EventType.SCROLL
        self.assertRaises(AttributeError, lambda: getattr(event, "foo_bar"))

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_scroll_event(self):
        event = Gdk.Event.new(Gdk.EventType.SCROLL)
        assert event.direction == Gdk.ScrollDirection.UP

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_event_strip_boolean(self):
        ev = Gdk.EventButton()
        ev.type = Gdk.EventType.BUTTON_PRESS
        assert ev.get_coords() == (0.0, 0.0)

        # https://gitlab.gnome.org/GNOME/pygobject/issues/85
        ev = Gdk.Event.new(Gdk.EventType.BUTTON_PRESS)
        assert ev.get_coords() == (True, 0.0, 0.0)

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_event_touch(self):
        event = Gdk.Event.new(Gdk.EventType.TOUCH_BEGIN)
        self.assertEqual(event.type, Gdk.EventType.TOUCH_BEGIN)

        # emulating_pointer is unique to touch events
        self.assertFalse(event.emulating_pointer)
        self.assertFalse(event.touch.emulating_pointer)

        event.emulating_pointer = True
        self.assertTrue(event.emulating_pointer)
        self.assertTrue(event.touch.emulating_pointer)

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_event_setattr(self):
        event = Gdk.Event.new(Gdk.EventType.DRAG_MOTION)
        event.x_root, event.y_root = 0, 5
        self.assertEqual(event.dnd.x_root, 0)
        self.assertEqual(event.dnd.y_root, 5)
        self.assertEqual(event.x_root, 0)
        self.assertEqual(event.y_root, 5)

        # this used to work, keep it that way
        self.assertFalse(hasattr(event, "foo_bar"))
        event.foo_bar = 42

        # unhandled type
        event.type = Gdk.EventType.EVENT_LAST
        with pytest.raises(AttributeError):
            event.foo_bar
        event.foo_bar = 42
        assert event.foo_bar == 42

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_event_repr(self):
        event = Gdk.Event.new(Gdk.EventType.CONFIGURE)
        self.assertTrue("CONFIGURE" in repr(event))

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_event_structures(self):
        def button_press_cb(button, event):
            self.assertTrue(isinstance(event, Gdk.EventButton))
            self.assertTrue(event.type == Gdk.EventType.BUTTON_PRESS)
            self.assertEqual(event.send_event, 0)
            self.assertEqual(event.get_state(), Gdk.ModifierType.CONTROL_MASK)
            self.assertEqual(event.get_root_coords(), (2, 5))

            event.time = 12345
            self.assertEqual(event.get_time(), 12345)

        w = Gtk.Window()
        b = Gtk.Button()
        b.connect("button-press-event", button_press_cb)
        w.add(b)
        b.show()
        b.realize()
        Gdk.test_simulate_button(
            b.get_window(),
            2,
            5,
            0,
            Gdk.ModifierType.CONTROL_MASK,
            Gdk.EventType.BUTTON_PRESS,
        )

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_cursor(self):
        self.assertEqual(Gdk.Cursor, gi.overrides.Gdk.Cursor)
        with capture_glib_deprecation_warnings():
            c = Gdk.Cursor(Gdk.CursorType.WATCH)
        self.assertNotEqual(c, None)
        with capture_glib_deprecation_warnings():
            c = Gdk.Cursor(cursor_type=Gdk.CursorType.WATCH)
        self.assertNotEqual(c, None)

        display_manager = Gdk.DisplayManager.get()
        display = display_manager.get_default_display()

        test_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, False, 8, 5, 10)

        with capture_glib_deprecation_warnings() as warn:
            c = Gdk.Cursor(display, test_pixbuf, y=0, x=0)
            self.assertNotEqual(c, None)

            self.assertEqual(len(warn), 1)
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))
            self.assertRegex(str(warn[0].message), ".*new_from_pixbuf.*")

        self.assertRaises(ValueError, Gdk.Cursor, 1, 2, 3)

        with capture_glib_deprecation_warnings() as warn:
            c = Gdk.Cursor(display, Gdk.CursorType.WATCH)
        assert len(warn) == 1
        # on macOS the type is switched to PIXMAP behind the scenes
        assert c.props.cursor_type in (
            Gdk.CursorType.WATCH,
            Gdk.CursorType.CURSOR_IS_PIXMAP,
        )
        assert c.props.display == display

    @unittest.skipUnless(GDK4, "only gdk4")
    def test_cursor_gdk4(self):
        Gdk.Cursor()
        Gdk.Cursor(name="foo")
        Gdk.Cursor(fallback=Gdk.Cursor())

    def test_flags(self):
        self.assertEqual(Gdk.ModifierType.META_MASK | 0, 0x10000000)
        self.assertEqual(hex(Gdk.ModifierType.META_MASK), "0x10000000")
        self.assertEqual(
            repr(Gdk.ModifierType.META_MASK), "<ModifierType.META_MASK: 268435456>"
        )

        # RELEASE_MASK does not exist in gdk4
        if not GDK4:
            self.assertEqual(Gdk.ModifierType.RELEASE_MASK | 0, 0x40000000)
            self.assertEqual(hex(Gdk.ModifierType.RELEASE_MASK), "0x40000000")
            self.assertEqual(
                repr(Gdk.ModifierType.RELEASE_MASK),
                "<ModifierType.RELEASE_MASK: 1073741824>",
            )

            self.assertEqual(
                Gdk.ModifierType.RELEASE_MASK | Gdk.ModifierType.META_MASK, 0x50000000
            )
            self.assertIn(
                repr(Gdk.ModifierType.RELEASE_MASK | Gdk.ModifierType.META_MASK),
                {
                    "<ModifierType.META_MASK|RELEASE_MASK: 1342177280>",
                    "<ModifierType.RELEASE_MASK|META_MASK: 1342177280>",
                },
            )

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_color_parse(self):
        with capture_glib_deprecation_warnings():
            c = Gdk.color_parse("#00FF80")
        self.assertEqual(c.red, 0)
        self.assertEqual(c.green, 65535)
        self.assertEqual(c.blue, 32896)
        self.assertEqual(Gdk.color_parse("bogus"), None)

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_color_representations(self):
        # __repr__ should generate a string which is parsable when possible
        # http://docs.python.org/2/reference/datamodel.html#object.__repr__
        color = Gdk.Color(red=65535, green=32896, blue=1)
        self.assertEqual(eval(repr(color)), color)

    def test_rgba_representations(self):
        rgba = Gdk.RGBA(red=1.0, green=0.8, blue=0.6, alpha=0.4)
        self.assertEqual(eval(repr(rgba)), rgba)

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_rectangle_functions(self):
        # https://bugzilla.gnome.org/show_bug.cgi?id=756364
        a = Gdk.Rectangle()
        b = Gdk.Rectangle()
        self.assertTrue(isinstance(Gdk.rectangle_union(a, b), Gdk.Rectangle))
        intersect, rect = Gdk.rectangle_intersect(a, b)
        self.assertTrue(isinstance(rect, Gdk.Rectangle))
        self.assertTrue(isinstance(intersect, bool))

    @unittest.skipIf(GDK4, "not in gdk4")
    def test_atom_repr_str(self):
        atom = Gdk.atom_intern("", True)
        assert re.match(r"<Gdk.Atom\(\d+\)>", repr(atom))
        assert re.match(r"Gdk.Atom<\d+>", str(atom))

    @unittest.skipUnless(GDK4, "only in gdk4")
    @unittest.skipUnless(gtkver() >= (4, 8, 0), "constructor available since 4.8")
    def test_file_list(self):
        f = Gio.File.new_for_path("/tmp")
        filelist = Gdk.FileList([f])
        self.assertTrue(isinstance(filelist, Gdk.FileList))
        self.assertEqual(len(filelist), 1)
        self.assertEqual(filelist[0], f)

    @unittest.skipUnless(GDK4, "only in gdk4")
    @unittest.skipUnless(gtkver() >= (4, 21, 0), "available since 4.21")
    def test_paintable_flags(self):
        self.assertEqual(Gdk.PaintableFlags.SIZE, Gdk.PaintableFlags.STATIC_SIZE)
        self.assertEqual(
            Gdk.PaintableFlags.CONTENTS, Gdk.PaintableFlags.STATIC_CONTENTS
        )



# test_error.py: Tests for GError wrapper implementation
#
# Copyright (C) 2012 Will Thompson
# Copyright (C) 2013 Martin Pitt
# Copyright (C) 2014 Simon Feltman <sfeltman@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import unittest
import pickle

from gi.repository import GLib
from gi.repository import GIMarshallingTests


class TestType(unittest.TestCase):
    def test_attributes(self):
        e = GLib.Error("test message", "mydomain", 42)
        self.assertEqual(e.message, "test message")
        self.assertEqual(e.domain, "mydomain")
        self.assertEqual(e.code, 42)

    def test_new_literal(self):
        mydomain = GLib.quark_from_string("mydomain")
        e = GLib.Error.new_literal(mydomain, "test message", 42)
        self.assertEqual(e.message, "test message")
        self.assertEqual(e.domain, "mydomain")
        self.assertEqual(e.code, 42)

    def test_matches(self):
        mydomain = GLib.quark_from_string("mydomain")
        notmydomain = GLib.quark_from_string("notmydomain")
        e = GLib.Error("test message", "mydomain", 42)
        self.assertTrue(e.matches(mydomain, 42))
        self.assertFalse(e.matches(notmydomain, 42))
        self.assertFalse(e.matches(mydomain, 40))

    def test_str(self):
        e = GLib.Error("test message", "mydomain", 42)
        self.assertEqual(str(e), "mydomain: test message (42)")

    def test_repr(self):
        e = GLib.Error("test message", "mydomain", 42)
        self.assertEqual(repr(e), "GLib.Error('test message', 'mydomain', 42)")

    def test_inheritance(self):
        self.assertTrue(issubclass(GLib.Error, RuntimeError))

    def test_pickle(self):
        def check_pickle(e):
            assert isinstance(e, GLib.Error)
            new_e = pickle.loads(pickle.dumps(e))
            assert type(new_e) is type(e)
            assert repr(e) == repr(new_e)

        e = GLib.Error("test message", "mydomain", 42)
        check_pickle(e)

        try:
            GLib.file_get_contents("")
        except Exception as e:
            check_pickle(e)


class ObjectWithVFuncException(GIMarshallingTests.Object):
    def do_vfunc_meth_with_err(self, x):
        if x == 42:
            return True

        msg = f"unexpected value {x:d}"
        raise GLib.Error(msg, "mydomain", 42)


class TestMarshalling(unittest.TestCase):
    def test_array_in_crash(self):
        # Previously there was a bug in invoke, in which C arrays were unwrapped
        # from inside GArrays to be passed to the C function. But when a GError was
        # set, invoke would attempt to free the C array as if it were a GArray.
        # This crash is only for C arrays. It does not happen for C functions which
        # take in GArrays. See https://bugzilla.gnome.org/show_bug.cgi?id=642708
        self.assertRaises(GLib.Error, GIMarshallingTests.gerror_array_in, [1, 2, 3])

    def test_out(self):
        # See https://bugzilla.gnome.org/show_bug.cgi?id=666098
        error, debug = GIMarshallingTests.gerror_out()

        self.assertIsInstance(error, GLib.Error)
        self.assertEqual(error.domain, GIMarshallingTests.CONSTANT_GERROR_DOMAIN)
        self.assertEqual(error.code, GIMarshallingTests.CONSTANT_GERROR_CODE)
        self.assertEqual(error.message, GIMarshallingTests.CONSTANT_GERROR_MESSAGE)
        self.assertEqual(debug, GIMarshallingTests.CONSTANT_GERROR_DEBUG_MESSAGE)

    def test_out_transfer_none(self):
        # See https://bugzilla.gnome.org/show_bug.cgi?id=666098
        error, debug = GIMarshallingTests.gerror_out_transfer_none()

        self.assertIsInstance(error, GLib.Error)
        self.assertEqual(error.domain, GIMarshallingTests.CONSTANT_GERROR_DOMAIN)
        self.assertEqual(error.code, GIMarshallingTests.CONSTANT_GERROR_CODE)
        self.assertEqual(error.message, GIMarshallingTests.CONSTANT_GERROR_MESSAGE)
        self.assertEqual(GIMarshallingTests.CONSTANT_GERROR_DEBUG_MESSAGE, debug)

    def test_return(self):
        # See https://bugzilla.gnome.org/show_bug.cgi?id=666098
        error = GIMarshallingTests.gerror_return()

        self.assertIsInstance(error, GLib.Error)
        self.assertEqual(error.domain, GIMarshallingTests.CONSTANT_GERROR_DOMAIN)
        self.assertEqual(error.code, GIMarshallingTests.CONSTANT_GERROR_CODE)
        self.assertEqual(error.message, GIMarshallingTests.CONSTANT_GERROR_MESSAGE)

    def test_exception(self):
        with self.assertRaises(GLib.Error) as context:
            GIMarshallingTests.gerror()

        e = context.exception
        self.assertEqual(e.domain, GIMarshallingTests.CONSTANT_GERROR_DOMAIN)
        self.assertEqual(e.code, GIMarshallingTests.CONSTANT_GERROR_CODE)
        self.assertEqual(e.message, GIMarshallingTests.CONSTANT_GERROR_MESSAGE)

    def test_vfunc_no_exception(self):
        obj = ObjectWithVFuncException()
        self.assertTrue(obj.vfunc_meth_with_error(42))

    def test_vfunc_gerror_exception(self):
        obj = ObjectWithVFuncException()
        with self.assertRaises(GLib.Error) as context:
            obj.vfunc_meth_with_error(-1)

        e = context.exception
        self.assertEqual(e.message, "unexpected value -1")
        self.assertEqual(e.domain, "mydomain")
        self.assertEqual(e.code, 42)

    def tests_compare_two_gerrors_in_gvalue(self):
        error = GLib.Error.new_literal(1, "error", 1)
        error1 = GLib.Error.new_literal(1, "error", 1)

        GIMarshallingTests.compare_two_gerrors_in_gvalue(error, error1)


def test_nullable_error():
    error = GLib.Error.new_literal(1, "error", 1)

    assert GIMarshallingTests.nullable_gerror(error) == 1
    assert GIMarshallingTests.nullable_gerror(None) == 0



import os
import select
import signal
import unittest
import warnings

from gi.repository import GLib

from .helper import capture_exceptions


class TestMainLoop(unittest.TestCase):
    @unittest.skipUnless(hasattr(os, "fork"), "no os.fork available")
    def test_exception_handling(self):
        pipe_r, pipe_w = os.pipe()

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", ".*may lead to deadlock.*")
            pid = os.fork()
        if pid == 0:
            os.close(pipe_w)
            select.select([pipe_r], [], [])
            os.close(pipe_r)
            os._exit(1)

        def child_died(pid, status, loop):
            loop.quit()
            raise Exception("deadbabe")

        loop = GLib.MainLoop()
        GLib.child_watch_add(GLib.PRIORITY_DEFAULT, pid, child_died, loop)

        os.close(pipe_r)
        os.write(pipe_w, b"Y")
        os.close(pipe_w)

        with capture_exceptions() as exc:
            loop.run()

        assert len(exc) == 1
        assert exc[0].type is Exception
        assert exc[0].value.args[0] == "deadbabe"

    @unittest.skipUnless(hasattr(os, "fork"), "no os.fork available")
    @unittest.skipIf(os.environ.get("PYGI_TEST_GDB"), "SIGINT stops gdb")
    def test_sigint(self):
        r, w = os.pipe()
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", ".*may lead to deadlock.*")
            pid = os.fork()
        if pid == 0:
            # wait for the parent process loop to start
            os.read(r, 1)
            os.close(r)

            os.kill(os.getppid(), signal.SIGINT)
            os._exit(0)

        def notify_child():
            # tell the child that it can kill the parent
            os.write(w, b"X")
            os.close(w)

        GLib.idle_add(notify_child)
        loop = GLib.MainLoop()
        try:
            loop.run()
            self.fail("expected KeyboardInterrupt exception")
        except KeyboardInterrupt:
            pass
        self.assertFalse(loop.is_running())
        os.waitpid(pid, 0)



import pytest
import platform
import unittest

import asyncio
from gi.repository import GLib, Gio
from gi.events import GLibEventLoopPolicy


class TestAsync(unittest.TestCase):
    def setUp(self):
        policy = GLibEventLoopPolicy()
        asyncio.set_event_loop_policy(policy)
        self.addCleanup(asyncio.set_event_loop_policy, None)
        self.loop = policy.get_event_loop()
        self.addCleanup(self.loop.close)

    def test_async_enumerate(self):
        f = Gio.file_new_for_path("./")

        called = False

        def cb():
            nonlocal called
            called = True

        async def run():
            nonlocal called, self

            self.loop.call_soon(cb)

            iter_info = [
                info.get_name()
                for info in await f.enumerate_children_async(
                    "standard::*", 0, GLib.PRIORITY_DEFAULT
                )
            ]

            # The await runs the mainloop and cb is called.
            self.assertEqual(called, True)

            next_info = []
            enumerator = f.enumerate_children("standard::*", 0, None)
            while True:
                info = enumerator.next_file(None)
                if info is None:
                    break
                next_info.append(info.get_name())

            self.assertEqual(iter_info, next_info)

        self.loop.run_until_complete(run())

    def test_async_cancellation(self):
        """Cancellation raises G_IO_ERROR_CANCELLED."""
        f = Gio.file_new_for_path("./")

        async def run():
            nonlocal self

            # cancellable created implicitly
            res = f.enumerate_children_async("standard::*", 0, GLib.PRIORITY_DEFAULT)
            res.cancel()
            with self.assertRaisesRegex(GLib.GError, "Operation was cancelled"):
                await res

            # cancellable passed explicitly
            cancel = Gio.Cancellable()
            res = f.enumerate_children_async(
                "standard::*", 0, GLib.PRIORITY_DEFAULT, cancel
            )
            self.assertEqual(res.cancellable, cancel)
            cancel.cancel()
            with self.assertRaisesRegex(GLib.GError, "Operation was cancelled"):
                await res

        self.loop.run_until_complete(run())

    def test_not_completed(self):
        """Querying an uncompleted task raises exceptions."""
        f = Gio.file_new_for_path("./")

        async def run():
            nonlocal self

            # cancellable created implicitly
            res = f.enumerate_children_async("standard::*", 0, GLib.PRIORITY_DEFAULT)
            with self.assertRaises(asyncio.InvalidStateError):
                res.result()

            with self.assertRaises(asyncio.InvalidStateError):
                res.exception()

            # And, await it
            await res

        self.loop.run_until_complete(run())

    def test_async_cancel_completed(self):
        """Cancelling a completed task just cancels the cancellable."""
        f = Gio.file_new_for_path("./")

        async def run():
            nonlocal self

            res = f.enumerate_children_async("standard::*", 0, GLib.PRIORITY_DEFAULT)
            await res
            assert not res.cancellable.is_cancelled()
            res.cancel()
            assert res.cancellable.is_cancelled()

        self.loop.run_until_complete(run())

    def test_async_completed_add_cb(self):
        """Adding a done cb to a completed future queues it with call_soon."""
        f = Gio.file_new_for_path("./")

        called = False

        def cb():
            nonlocal called
            called = True

        async def run():
            nonlocal called, self

            res = f.enumerate_children_async("standard::*", 0, GLib.PRIORITY_DEFAULT)
            await res
            self.loop.call_soon(cb)

            # Python await is smart and does not iterate the EventLoop
            await res
            assert not called

            # So create a new future and wait on that
            fut = asyncio.Future()

            def done_cb(res):
                nonlocal fut
                fut.set_result(res.result())

            res.add_done_callback(done_cb)
            await fut
            assert called

        self.loop.run_until_complete(run())

    @pytest.mark.xfail(
        platform.python_implementation() == "PyPy",
        reason="Exception reporting does not work in pypy",
    )
    def test_deleting_failed_logs(self):
        f = Gio.file_new_for_path("./")

        async def run():
            nonlocal self

            res = f.enumerate_children_async("standard::*", 0, GLib.PRIORITY_DEFAULT)
            res.cancel()
            # Cancellation in Gio is not immediate, so sleep for a bit
            await asyncio.sleep(0.5)

        exc = None
        msg = None

        def handler(loop, context):
            nonlocal exc, msg
            msg = context["message"]
            exc = context["exception"]

        self.loop.set_exception_handler(handler)
        self.loop.run_until_complete(run())

        self.assertRegex(msg, ".*exception was never retrieved")
        self.assertIsInstance(exc, GLib.GError)
        assert exc.matches(Gio.io_error_quark(), Gio.IOErrorEnum.CANCELLED)

    def test_no_running_loop(self):
        f = Gio.file_new_for_path("./")

        res = f.enumerate_children_async("standard::*", 0, GLib.PRIORITY_DEFAULT)
        self.assertIsNone(res)

    def test_wrong_default_context(self):
        f = Gio.file_new_for_path("./")

        async def run():  # noqa: RUF029
            nonlocal self

            ctx = GLib.MainContext.new()
            GLib.MainContext.push_thread_default(ctx)
            try:
                res = f.enumerate_children_async(
                    "standard::*", 0, GLib.PRIORITY_DEFAULT
                )
                self.assertIsNone(res)
            finally:
                GLib.MainContext.pop_thread_default(ctx)

        self.loop.run_until_complete(run())



import unittest

from gi.repository import GLib

import testhelper


class TestThread(unittest.TestCase):
    def setUp(self):
        self.main = GLib.MainLoop()
        self.called = False

    def from_thread_cb(self, test, enum):
        assert test == self.obj
        assert int(enum) == 0
        assert type(enum) is not int
        self.called = True
        GLib.idle_add(self.timeout_cb)

    def idle_cb(self):
        self.obj = testhelper.get_test_thread()
        self.obj.connect("from-thread", self.from_thread_cb)
        self.obj.emit("emit-signal")

    def test_extension_module(self):
        GLib.idle_add(self.idle_cb)
        GLib.timeout_add(2000, self.timeout_cb)
        self.main.run()
        self.assertTrue(self.called)

    def timeout_cb(self):
        self.main.quit()



from gi.repository import GLib, Gio, Regress


def iteration():
    ctx = GLib.main_context_default()
    while ctx.pending():
        ctx.iteration(False)


def test_async_callback():
    result = []
    cancel = Gio.Cancellable()

    def callback(obj, res):
        result.append(obj.function_finish(res))

    obj = Regress.TestObj()
    obj.function_async(GLib.PRIORITY_DEFAULT, cancellable=cancel, callback=callback)
    obj.function_thaw_async()

    iteration()

    assert result == [True]


def test_async_callback_with_extra_callbacks():
    result = []

    def callback(obj, res):
        result.append(obj.function2_finish(res))

    obj = Regress.TestObj()
    obj.function2(GLib.PRIORITY_DEFAULT, None, None, None, callback)
    obj.function_thaw_async()

    iteration()

    assert result == [(True, True, None)]


def test_async_callback_with_extra_callbacks_filled_in():
    result = []
    cancel = Gio.Cancellable()

    def test_cb(data):
        result.append(data)
        return 0

    def callback(obj, res):
        result.append(obj.function2_finish(res))

    obj = Regress.TestObj()
    obj.function2(GLib.PRIORITY_DEFAULT, cancel, test_cb, ("test_cb data",), callback)
    obj.function_thaw_async()

    iteration()

    assert result == ["test_cb data", (True, True, None)]


def test_async_callback_with_extra_callbacks_as_kwarg():
    result = []
    cancel = Gio.Cancellable()

    def test_cb(data):
        result.append(data)
        return 0

    def callback(obj, res):
        result.append(obj.function2_finish(res))

    obj = Regress.TestObj()
    obj.function2(
        GLib.PRIORITY_DEFAULT, cancellable=cancel, test_cb=test_cb, callback=callback
    )
    obj.function_thaw_async()

    iteration()

    assert result == [None, (True, True, None)]


def test_async_callback_with_extra_callbacks_as_kwarg_and_user_data():
    result = []
    cancel = Gio.Cancellable()

    def test_cb(data):
        result.append(data)
        return 0

    def callback(obj, res, data):
        result.append(obj.function2_finish(res))

    obj = Regress.TestObj()
    obj.function2(
        GLib.PRIORITY_DEFAULT,
        cancellable=cancel,
        test_cb=test_cb,
        callback=callback,
        user_data=None,
    )
    obj.function_thaw_async()

    iteration()

    assert result == [None, (True, True, None)]



# test_typeclass.py: Tests for GTypeClass related methods and marshalling.
#
# Copyright (C) 2014 Simon Feltman <sfeltman@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import unittest

from gi.repository import GObject
from gi.repository import GIMarshallingTests


class TestCoercion(unittest.TestCase):
    def test_coerce_from_class(self):
        prop = GObject.ObjectClass.find_property(
            GIMarshallingTests.PropertiesObject, "some-int"
        )

        self.assertIsInstance(prop, GObject.ParamSpec)
        self.assertEqual(prop.name, "some-int")
        self.assertEqual(prop.value_type, GObject.TYPE_INT)
        self.assertEqual(prop.owner_type, GIMarshallingTests.PropertiesObject.__gtype__)

    def test_coerce_from_gtype(self):
        gtype = GIMarshallingTests.PropertiesObject.__gtype__
        prop = GObject.ObjectClass.find_property(gtype, "some-int")

        self.assertIsInstance(prop, GObject.ParamSpec)
        self.assertEqual(prop.name, "some-int")
        self.assertEqual(prop.value_type, GObject.TYPE_INT)
        self.assertEqual(prop.owner_type, gtype)

    def test_coerce_from_instance(self):
        obj = GIMarshallingTests.PropertiesObject()
        prop = GObject.ObjectClass.find_property(obj, "some-int")

        self.assertIsInstance(prop, GObject.ParamSpec)
        self.assertEqual(prop.name, "some-int")
        self.assertEqual(prop.value_type, GObject.TYPE_INT)
        self.assertEqual(prop.owner_type, obj.__gtype__)

    def test_marshalling_error(self):
        with self.assertRaises(TypeError):
            GObject.ObjectClass.find_property(object, "some-int")

        with self.assertRaises(TypeError):
            GObject.ObjectClass.find_property(42, "some-int")


class TestTypeClassMethodsMovedToClass(unittest.TestCase):
    def test_list_child_properties(self):
        pspecs = GIMarshallingTests.PropertiesObject.list_properties()
        pnames = [pspec.name for pspec in pspecs]
        self.assertTrue("some-int" in pnames)
        self.assertTrue("some-float" in pnames)
        self.assertTrue("some-char" in pnames)

    def test_find_child_property(self):
        pspec = GIMarshallingTests.PropertiesObject.find_property("some-int")
        self.assertEqual(pspec.name, "some-int")



import tomllib

from pyproject_metadata import StandardMetadata


def update_metadata_in():
    with open("pyproject.toml", "rb") as fd:
        pyproject = tomllib.load(fd)

    metadata = StandardMetadata.from_pyproject(pyproject, allow_extra_keys=False)
    metadata.version = "@VERSION@"
    del metadata.readme
    del metadata.dependencies[:]

    with open("METADATA.in", "wb") as fd:
        fd.write(metadata.as_rfc822().as_bytes())


if __name__ == "__main__":
    update_metadata_in()



# Copyright 2017 Christoph Reiter
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.


import os
import signal
import unittest
import threading
from contextlib import contextmanager

try:
    from gi.repository import Gtk

    Gtk_version = Gtk._version
except ImportError:
    Gtk = None
    Gtk_version = None
from gi.repository import Gio, GLib
from gi._ossighelper import wakeup_on_signal, register_sigint_fallback


class TestOverridesWakeupOnAlarm(unittest.TestCase):
    @contextmanager
    def _run_with_timeout(self, timeout, abort_func):
        failed = []

        def fail():
            abort_func()
            failed.append(1)
            return True

        fail_id = GLib.timeout_add(timeout, fail)
        try:
            yield
        finally:
            GLib.source_remove(fail_id)
        self.assertFalse(failed)

    def test_basic(self):
        self.assertEqual(signal.set_wakeup_fd(-1), -1)
        with wakeup_on_signal():
            pass
        self.assertEqual(signal.set_wakeup_fd(-1), -1)

    def test_in_thread(self):
        failed = []

        def target():
            try:
                with wakeup_on_signal():
                    pass
            except ValueError or EnvironmentError:
                failed.append(1)

        t = threading.Thread(target=target)
        t.start()
        t.join(5)
        self.assertFalse(failed)

    @unittest.skipIf(os.name == "nt", "not on Windows")
    def test_glib_mainloop(self):
        loop = GLib.MainLoop()
        signal.signal(signal.SIGALRM, lambda *args: loop.quit())
        GLib.idle_add(signal.setitimer, signal.ITIMER_REAL, 0.001)

        with self._run_with_timeout(2000, loop.quit):
            loop.run()

    @unittest.skipIf(os.name == "nt", "not on Windows")
    def test_gio_application(self):
        app = Gio.Application()
        signal.signal(signal.SIGALRM, lambda *args: app.quit())
        GLib.idle_add(signal.setitimer, signal.ITIMER_REAL, 0.001)

        with self._run_with_timeout(2000, app.quit):
            app.hold()
            app.connect("activate", lambda *args: None)
            app.run()

    @unittest.skipIf(Gtk is None or os.name == "nt", "not on Windows")
    @unittest.skipIf(Gtk is None or Gtk_version == "4.0", "not in gtk4")
    def test_gtk_main(self):
        signal.signal(signal.SIGALRM, lambda *args: Gtk.main_quit())
        GLib.idle_add(signal.setitimer, signal.ITIMER_REAL, 0.001)

        with self._run_with_timeout(2000, Gtk.main_quit):
            Gtk.main()

    @unittest.skipIf(Gtk is None or os.name == "nt", "not on Windows")
    @unittest.skipIf(Gtk is None or Gtk_version == "4.0", "not in gtk4")
    def test_gtk_dialog_run(self):
        w = Gtk.Window()
        d = Gtk.Dialog(transient_for=w)
        signal.signal(signal.SIGALRM, lambda *args: d.destroy())
        GLib.idle_add(signal.setitimer, signal.ITIMER_REAL, 0.001)

        with self._run_with_timeout(2000, d.destroy):
            d.run()


class TestSigintFallback(unittest.TestCase):
    def setUp(self):
        self.assertEqual(signal.getsignal(signal.SIGINT), signal.default_int_handler)

    def tearDown(self):
        self.assertEqual(signal.getsignal(signal.SIGINT), signal.default_int_handler)

    def test_replace_handler_and_restore_nested(self):
        with register_sigint_fallback(lambda: None):
            new_handler = signal.getsignal(signal.SIGINT)
            self.assertNotEqual(new_handler, signal.default_int_handler)
            with register_sigint_fallback(lambda: None):
                self.assertTrue(signal.getsignal(signal.SIGINT) is new_handler)
        self.assertEqual(signal.getsignal(signal.SIGINT), signal.default_int_handler)

    def test_no_replace_if_not_default(self):
        def new_handler(*args):
            return None

        signal.signal(signal.SIGINT, new_handler)
        try:
            with register_sigint_fallback(lambda: None):
                self.assertTrue(signal.getsignal(signal.SIGINT) is new_handler)
                with register_sigint_fallback(lambda: None):
                    self.assertTrue(signal.getsignal(signal.SIGINT) is new_handler)
            self.assertTrue(signal.getsignal(signal.SIGINT) is new_handler)
        finally:
            signal.signal(signal.SIGINT, signal.default_int_handler)

    def test_noop_in_threads(self):
        failed = []

        def target():
            try:
                with (
                    register_sigint_fallback(lambda: None),
                    register_sigint_fallback(lambda: None),
                ):
                    self.assertTrue(
                        signal.getsignal(signal.SIGINT) is signal.default_int_handler
                    )
            except:
                failed.append(1)

        t = threading.Thread(target=target)
        t.start()
        t.join(5)
        self.assertFalse(failed)

    @unittest.skipIf(os.name == "nt", "not on Windows")
    def test_no_replace_if_set_by_glib(self):
        id_ = GLib.unix_signal_add(
            GLib.PRIORITY_DEFAULT, signal.SIGINT, lambda *args: None
        )
        try:
            # signal.getsignal() doesn't pick up that unix_signal_add()
            # has changed the handler, but we should anyway.
            self.assertEqual(
                signal.getsignal(signal.SIGINT), signal.default_int_handler
            )
            with register_sigint_fallback(lambda: None):
                self.assertEqual(
                    signal.getsignal(signal.SIGINT), signal.default_int_handler
                )
            self.assertEqual(
                signal.getsignal(signal.SIGINT), signal.default_int_handler
            )
        finally:
            GLib.source_remove(id_)
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            signal.signal(signal.SIGINT, signal.default_int_handler)



import sys
import gc
import unittest
import warnings
import weakref
import platform

import pytest

from gi.repository import GObject, GLib, Gio
from gi import PyGIDeprecationWarning
from gi.module import get_introspection_module
from gi import _gi

import testhelper
from .helper import capture_glib_deprecation_warnings


@pytest.mark.skipif(platform.python_implementation() == "PyPy", reason="crashes")
def test_gobject_weak_ref():
    called = []

    def callback(*args):
        called.extend(args)

    # object gets finalized
    obj = GObject.Object()
    obj.weak_ref(callback, 1)
    del obj
    gc.collect()
    assert called == [1]
    del called[:]

    # wrapper gets finalized first
    obj = GObject.Object()
    pyref = weakref.ref(obj, lambda x: callback(-2))
    value = GObject.Value(GObject.Object, obj)
    ref = obj.weak_ref(callback, 2)
    del obj
    gc.collect()
    assert called == [-2]
    del pyref
    value.unset()
    gc.collect()
    assert called == [-2, 2]
    del called[:]

    # weakref gets unregistered first
    obj = GObject.Object()
    ref = obj.weak_ref(callback, 3)
    ref.unref()
    del obj
    gc.collect()
    assert not called

    # weakref gets GCed
    obj = GObject.Object()
    obj.weak_ref(callback, 4)
    gc.collect()
    del obj
    assert called == [4]


class TestGObjectAPI(unittest.TestCase):
    def test_run_dispose(self):
        class TestObject(GObject.GObject):
            int_prop = GObject.Property(default=0, type=int)

        obj = TestObject()
        called = []

        def on_notify(*args):
            called.append(args)

        obj.connect("notify::int-prop", on_notify)
        obj.notify("int-prop")
        obj.notify("int-prop")
        # after this everything should be disconnected
        obj.run_dispose()
        obj.notify("int-prop")
        obj.notify("int-prop")
        assert len(called) == 2

    def test_call_method_uninitialized_instance(self):
        obj = GObject.Object.__new__(GObject.Object)
        with self.assertRaisesRegex(RuntimeError, ".*is not initialized"):
            obj.notify("foo")

    def test_gobject_inheritance(self):
        # GObject.Object is a class hierarchy as follows:
        # overrides.Object -> introspection.Object -> static.GObject
        GIObjectModule = get_introspection_module("GObject")
        self.assertTrue(issubclass(GObject.Object, GIObjectModule.Object))
        self.assertTrue(issubclass(GIObjectModule.Object, _gi.GObject))

        self.assertEqual(_gi.GObject.__gtype__, GObject.TYPE_OBJECT)
        self.assertEqual(GIObjectModule.Object.__gtype__, GObject.TYPE_OBJECT)
        self.assertEqual(GObject.Object.__gtype__, GObject.TYPE_OBJECT)

        # The pytype wrapper should hold the outer most Object class from overrides.
        self.assertEqual(GObject.TYPE_OBJECT.pytype, GObject.Object)

    def test_gobject_unsupported_overrides(self):
        obj = GObject.Object()

        with self.assertRaisesRegex(
            RuntimeError, "Data access methods are unsupported.*"
        ):
            obj.get_data()

        with self.assertRaisesRegex(
            RuntimeError, "This method is currently unsupported.*"
        ):
            obj.force_floating()

    def test_compat_api(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # GObject formerly exposed a lot of GLib's functions
            self.assertEqual(GObject.markup_escape_text("foo"), "foo")

            ml = GObject.MainLoop()
            self.assertFalse(ml.is_running())

            context = GObject.main_context_default()
            self.assertTrue(context.pending() in [False, True])

            context = GObject.MainContext()
            self.assertFalse(context.pending())

            self.assertTrue(issubclass(w[0].category, PyGIDeprecationWarning))
            self.assertTrue("GLib.markup_escape_text" in str(w[0]), str(w[0]))

            self.assertLess(GObject.PRIORITY_HIGH, GObject.PRIORITY_DEFAULT)

    def test_min_max_int(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)

            self.assertEqual(GObject.G_MAXINT16, 2**15 - 1)
            self.assertEqual(GObject.G_MININT16, -(2**15))
            self.assertEqual(GObject.G_MAXUINT16, 2**16 - 1)

            self.assertEqual(GObject.G_MAXINT32, 2**31 - 1)
            self.assertEqual(GObject.G_MININT32, -(2**31))
            self.assertEqual(GObject.G_MAXUINT32, 2**32 - 1)

            self.assertEqual(GObject.G_MAXINT64, 2**63 - 1)
            self.assertEqual(GObject.G_MININT64, -(2**63))
            self.assertEqual(GObject.G_MAXUINT64, 2**64 - 1)


class TestReferenceCounting(unittest.TestCase):
    def test_regular_object(self):
        obj = GObject.GObject()
        self.assertEqual(obj.__grefcount__, 1)

        obj = GObject.new(GObject.GObject)
        self.assertEqual(obj.__grefcount__, 1)

    def test_floating(self):
        obj = testhelper.Floating()
        self.assertEqual(obj.__grefcount__, 1)

        obj = GObject.new(testhelper.Floating)
        self.assertEqual(obj.__grefcount__, 1)

    def test_owned_by_library(self):
        # Upon creation, the refcount of the object should be 2:
        # - someone already has a reference on the new object.
        # - the python wrapper should hold its own reference.
        obj = testhelper.OwnedByLibrary()
        self.assertEqual(obj.__grefcount__, 2)

        # We ask the library to release its reference, so the only
        # remaining ref should be our wrapper's. Once the wrapper
        # will run out of scope, the object will get finalized.
        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_owned_by_library_out_of_scope(self):
        obj = testhelper.OwnedByLibrary()
        self.assertEqual(obj.__grefcount__, 2)

        # We are manually taking the object out of scope. This means
        # that our wrapper has been freed, and its reference dropped. We
        # cannot check it but the refcount should now be 1 (the ref held
        # by the library is still there, we didn't call release()
        obj = None

        # When we get the object back from the lib, the wrapper is
        # re-created, so our refcount will be 2 once again.
        obj = testhelper.owned_by_library_get_instance_list()[0]
        self.assertEqual(obj.__grefcount__, 2)

        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_owned_by_library_using_gobject_new(self):
        # Upon creation, the refcount of the object should be 2:
        # - someone already has a reference on the new object.
        # - the python wrapper should hold its own reference.
        obj = GObject.new(testhelper.OwnedByLibrary)
        self.assertEqual(obj.__grefcount__, 2)

        # We ask the library to release its reference, so the only
        # remaining ref should be our wrapper's. Once the wrapper
        # will run out of scope, the object will get finalized.
        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_owned_by_library_out_of_scope_using_gobject_new(self):
        obj = GObject.new(testhelper.OwnedByLibrary)
        self.assertEqual(obj.__grefcount__, 2)

        # We are manually taking the object out of scope. This means
        # that our wrapper has been freed, and its reference dropped. We
        # cannot check it but the refcount should now be 1 (the ref held
        # by the library is still there, we didn't call release()
        obj = None

        # When we get the object back from the lib, the wrapper is
        # re-created, so our refcount will be 2 once again.
        obj = testhelper.owned_by_library_get_instance_list()[0]
        self.assertEqual(obj.__grefcount__, 2)

        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_floating_and_sunk(self):
        # Upon creation, the refcount of the object should be 2:
        # - someone already has a reference on the new object.
        # - the python wrapper should hold its own reference.
        obj = testhelper.FloatingAndSunk()
        self.assertEqual(obj.__grefcount__, 2)

        # We ask the library to release its reference, so the only
        # remaining ref should be our wrapper's. Once the wrapper
        # will run out of scope, the object will get finalized.
        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_floating_and_sunk_out_of_scope(self):
        obj = testhelper.FloatingAndSunk()
        self.assertEqual(obj.__grefcount__, 2)

        # We are manually taking the object out of scope. This means
        # that our wrapper has been freed, and its reference dropped. We
        # cannot check it but the refcount should now be 1 (the ref held
        # by the library is still there, we didn't call release()
        obj = None

        # When we get the object back from the lib, the wrapper is
        # re-created, so our refcount will be 2 once again.
        obj = testhelper.floating_and_sunk_get_instance_list()[0]
        self.assertEqual(obj.__grefcount__, 2)

        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_floating_and_sunk_using_gobject_new(self):
        # Upon creation, the refcount of the object should be 2:
        # - someone already has a reference on the new object.
        # - the python wrapper should hold its own reference.
        obj = GObject.new(testhelper.FloatingAndSunk)
        self.assertEqual(obj.__grefcount__, 2)

        # We ask the library to release its reference, so the only
        # remaining ref should be our wrapper's. Once the wrapper
        # will run out of scope, the object will get finalized.
        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_floating_and_sunk_out_of_scope_using_gobject_new(self):
        obj = GObject.new(testhelper.FloatingAndSunk)
        self.assertEqual(obj.__grefcount__, 2)

        # We are manually taking the object out of scope. This means
        # that our wrapper has been freed, and its reference dropped. We
        # cannot check it but the refcount should now be 1 (the ref held
        # by the library is still there, we didn't call release()
        obj = None

        # When we get the object back from the lib, the wrapper is
        # re-created, so our refcount will be 2 once again.
        obj = testhelper.floating_and_sunk_get_instance_list()[0]
        self.assertEqual(obj.__grefcount__, 2)

        obj.release()
        self.assertEqual(obj.__grefcount__, 1)

    def test_uninitialized_object(self):
        class Obj(GObject.GObject):
            def __init__(self):
                x = self.__grefcount__
                super().__init__()
                assert x >= 0

        # Accessing __grefcount__ before the object is initialized is wrong.
        # Ensure we get a proper exception instead of a crash.
        self.assertRaises(TypeError, Obj)


class A(GObject.GObject):
    def __init__(self):
        super().__init__()


class TestPythonReferenceCounting(unittest.TestCase):
    # Newly created instances should always have two references: one for
    # the GC, and one for the bound variable in the local scope.

    def test_new_instance_has_two_refs(self):
        obj = GObject.GObject()
        if hasattr(sys, "getrefcount"):
            self.assertEqual(
                sys.getrefcount(obj), 2 if sys.version_info < (3, 14) else 1
            )

    def test_new_instance_has_two_refs_using_gobject_new(self):
        obj = GObject.new(GObject.GObject)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(
                sys.getrefcount(obj), 2 if sys.version_info < (3, 14) else 1
            )

    def test_new_subclass_instance_has_two_refs(self):
        obj = A()
        if hasattr(sys, "getrefcount"):
            self.assertEqual(
                sys.getrefcount(obj), 2 if sys.version_info < (3, 14) else 1
            )

    def test_new_subclass_instance_has_two_refs_using_gobject_new(self):
        obj = GObject.new(A)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(
                sys.getrefcount(obj), 2 if sys.version_info < (3, 14) else 1
            )


class TestContextManagers(unittest.TestCase):
    class ContextTestObject(GObject.GObject):
        prop = GObject.Property(default=0, type=int)

    def on_prop_set(self, obj, prop):
        # Handler which tracks property changed notifications.
        self.tracking.append(obj.get_property(prop.name))

    def setUp(self):
        self.tracking = []
        self.obj = self.ContextTestObject()
        self.handler = self.obj.connect("notify::prop", self.on_prop_set)

    def test_freeze_notify_context(self):
        # Verify prop tracking list
        self.assertEqual(self.tracking, [])
        self.obj.props.prop = 1
        self.assertEqual(self.tracking, [1])
        self.obj.props.prop = 2
        self.assertEqual(self.tracking, [1, 2])
        self.assertEqual(self.obj.__grefcount__, 1)

        if hasattr(sys, "getrefcount"):
            pyref_count = sys.getrefcount(self.obj)

        # Using the context manager the tracking list should not be affected.
        # The GObject reference count should stay the same and the python
        # object ref-count should go up.
        with self.obj.freeze_notify():
            self.assertEqual(self.obj.__grefcount__, 1)
            if hasattr(sys, "getrefcount"):
                self.assertEqual(sys.getrefcount(self.obj), pyref_count + 1)
            self.obj.props.prop = 3
            self.assertEqual(self.obj.props.prop, 3)
            self.assertEqual(self.tracking, [1, 2])

        # After the context manager, the prop should have been modified,
        # the tracking list will be modified, and the python object ref
        # count goes back down.
        gc.collect()
        self.assertEqual(self.obj.props.prop, 3)
        self.assertEqual(self.tracking, [1, 2, 3])
        self.assertEqual(self.obj.__grefcount__, 1)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(self.obj), pyref_count)

    def test_handler_block_context(self):
        # Verify prop tracking list
        self.assertEqual(self.tracking, [])
        self.obj.props.prop = 1
        self.assertEqual(self.tracking, [1])
        self.obj.props.prop = 2
        self.assertEqual(self.tracking, [1, 2])
        self.assertEqual(self.obj.__grefcount__, 1)

        if hasattr(sys, "getrefcount"):
            pyref_count = sys.getrefcount(self.obj)

        # Using the context manager the tracking list should not be affected.
        # The GObject reference count should stay the same and the python
        # object ref-count should go up.
        with self.obj.handler_block(self.handler):
            self.assertEqual(self.obj.__grefcount__, 1)
            if hasattr(sys, "getrefcount"):
                self.assertEqual(sys.getrefcount(self.obj), pyref_count + 1)
            self.obj.props.prop = 3
            self.assertEqual(self.obj.props.prop, 3)
            self.assertEqual(self.tracking, [1, 2])

        # After the context manager, the prop should have been modified
        # the tracking list should have stayed the same and the GObject ref
        # count goes back down.
        gc.collect()
        self.assertEqual(self.obj.props.prop, 3)
        self.assertEqual(self.tracking, [1, 2])
        self.assertEqual(self.obj.__grefcount__, 1)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(self.obj), pyref_count)

    def test_freeze_notify_context_nested(self):
        self.assertEqual(self.tracking, [])
        with self.obj.freeze_notify():
            self.obj.props.prop = 1
            self.assertEqual(self.tracking, [])

            with self.obj.freeze_notify():
                self.obj.props.prop = 2
                self.assertEqual(self.tracking, [])

                with self.obj.freeze_notify():
                    self.obj.props.prop = 3
                    self.assertEqual(self.tracking, [])
                self.assertEqual(self.tracking, [])
            self.assertEqual(self.tracking, [])

        # Finally after last context, the notifications should have collapsed
        # and the last one sent.
        self.assertEqual(self.tracking, [3])

    def test_handler_block_context_nested(self):
        self.assertEqual(self.tracking, [])
        with self.obj.handler_block(self.handler):
            self.obj.props.prop = 1
            self.assertEqual(self.tracking, [])

            with self.obj.handler_block(self.handler):
                self.obj.props.prop = 2
                self.assertEqual(self.tracking, [])

                with self.obj.handler_block(self.handler):
                    self.obj.props.prop = 3
                    self.assertEqual(self.tracking, [])
                self.assertEqual(self.tracking, [])
            self.assertEqual(self.tracking, [])

        # Finally after last context, the notifications should have collapsed
        # and the last one sent.
        self.assertEqual(self.obj.props.prop, 3)
        self.assertEqual(self.tracking, [])

    def test_freeze_notify_normal_usage_ref_counts(self):
        # Ensure ref counts without using methods as context managers
        # maintain the same count.
        self.assertEqual(self.obj.__grefcount__, 1)
        self.obj.freeze_notify()
        self.assertEqual(self.obj.__grefcount__, 1)
        self.obj.thaw_notify()
        self.assertEqual(self.obj.__grefcount__, 1)

    def test_handler_block_normal_usage_ref_counts(self):
        self.assertEqual(self.obj.__grefcount__, 1)
        self.obj.handler_block(self.handler)
        self.assertEqual(self.obj.__grefcount__, 1)
        self.obj.handler_unblock(self.handler)
        self.assertEqual(self.obj.__grefcount__, 1)

    def test_freeze_notify_context_error(self):
        # Test an exception occurring within a freeze context exits the context
        try:
            with self.obj.freeze_notify():
                self.obj.props.prop = 1
                self.assertEqual(self.tracking, [])
                raise ValueError("Simulation")
        except ValueError:
            pass

        # Verify the property set within the context called notify.
        self.assertEqual(self.obj.props.prop, 1)
        self.assertEqual(self.tracking, [1])

        # Verify we are still not in a frozen context.
        self.obj.props.prop = 2
        self.assertEqual(self.tracking, [1, 2])

    def test_handler_block_context_error(self):
        # Test an exception occurring within a handler block exits the context
        try:
            with self.obj.handler_block(self.handler):
                self.obj.props.prop = 1
                self.assertEqual(self.tracking, [])
                raise ValueError("Simulation")
        except ValueError:
            pass

        # Verify the property set within the context didn't call notify.
        self.assertEqual(self.obj.props.prop, 1)
        self.assertEqual(self.tracking, [])

        # Verify we are still not in a handler block context.
        self.obj.props.prop = 2
        self.assertEqual(self.tracking, [2])


class TestPropertyBindings(unittest.TestCase):
    class TestObject(GObject.GObject):
        int_prop = GObject.Property(default=0, type=int)

    def setUp(self):
        self.source = self.TestObject()
        self.target = self.TestObject()

    def test_default_binding(self):
        binding = self.source.bind_property(
            "int_prop", self.target, "int_prop", GObject.BindingFlags.DEFAULT
        )
        binding = binding

        # Test setting value on source gets pushed to target
        self.source.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 1)

        # Test setting value on target does not change source
        self.target.props.int_prop = 2
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 2)

    def test_call_binding(self):
        binding = self.source.bind_property(
            "int_prop", self.target, "int_prop", GObject.BindingFlags.DEFAULT
        )
        with capture_glib_deprecation_warnings() as warn:
            result = binding()
        assert len(warn)
        assert result is binding

    def test_bidirectional_binding(self):
        binding = self.source.bind_property(
            "int_prop", self.target, "int_prop", GObject.BindingFlags.BIDIRECTIONAL
        )
        binding = binding

        # Test setting value on source gets pushed to target
        self.source.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 1)

        # Test setting value on target also changes source
        self.target.props.int_prop = 2
        self.assertEqual(self.source.int_prop, 2)
        self.assertEqual(self.target.int_prop, 2)

    def test_transform_to_only(self):
        def transform_to(binding, value, user_data=None):
            self.assertEqual(user_data, "test-data")
            return value * 2

        binding = self.source.bind_property(
            "int_prop",
            self.target,
            "int_prop",
            GObject.BindingFlags.DEFAULT,
            transform_to,
            user_data="test-data",
        )
        binding = binding

        self.source.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 2)

        self.target.props.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 1)

    def test_transform_from_only(self):
        def transform_from(binding, value, user_data=None):
            self.assertEqual(user_data, None)
            return value * 2

        binding = self.source.bind_property(
            "int_prop",
            self.target,
            "int_prop",
            GObject.BindingFlags.BIDIRECTIONAL,
            transform_from=transform_from,
        )
        binding = binding

        self.source.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 1)

        self.target.props.int_prop = 1
        self.assertEqual(self.source.int_prop, 2)
        self.assertEqual(self.target.int_prop, 1)

    def test_transform_bidirectional(self):
        test_data = object()

        def transform_to(binding, value, user_data=None):
            self.assertEqual(user_data, test_data)
            return value * 2

        def transform_from(binding, value, user_data=None):
            self.assertEqual(user_data, test_data)
            return value // 2

        if hasattr(sys, "getrefcount"):
            test_data_ref_count = sys.getrefcount(test_data)
            transform_to_ref_count = sys.getrefcount(transform_to)
            transform_from_ref_count = sys.getrefcount(transform_from)

        # bidirectional bindings
        binding = self.source.bind_property(
            "int_prop",
            self.target,
            "int_prop",
            GObject.BindingFlags.BIDIRECTIONAL,
            transform_to,
            transform_from,
            test_data,
        )
        binding = binding
        if hasattr(sys, "getrefcount"):
            binding_ref_count = sys.getrefcount(binding)
            binding_gref_count = binding.__grefcount__

        self.source.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 2)

        self.target.props.int_prop = 4
        self.assertEqual(self.source.int_prop, 2)
        self.assertEqual(self.target.int_prop, 4)

        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(binding), binding_ref_count)
            self.assertEqual(binding.__grefcount__, binding_gref_count)

        # test_data ref count increases by 2, once for each callback.
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(test_data), test_data_ref_count + 2)
            self.assertEqual(sys.getrefcount(transform_to), transform_to_ref_count + 1)
            self.assertEqual(
                sys.getrefcount(transform_from), transform_from_ref_count + 1
            )

        # Unbind should clear out the binding and its transforms
        binding.unbind()

        # Setting source or target should not change the other.
        self.target.int_prop = 3
        self.source.int_prop = 5
        self.assertEqual(self.target.int_prop, 3)
        self.assertEqual(self.source.int_prop, 5)

        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(test_data), test_data_ref_count)
            self.assertEqual(sys.getrefcount(transform_to), transform_to_ref_count)
            self.assertEqual(sys.getrefcount(transform_from), transform_from_ref_count)

    def test_explicit_unbind_clears_connection(self):
        self.assertEqual(self.source.int_prop, 0)
        self.assertEqual(self.target.int_prop, 0)

        # Test deleting binding reference removes binding.
        binding = self.source.bind_property("int_prop", self.target, "int_prop")
        self.source.int_prop = 1
        self.assertEqual(self.source.int_prop, 1)
        self.assertEqual(self.target.int_prop, 1)

        # unbind should clear out the bindings self reference
        binding.unbind()
        self.assertEqual(binding.__grefcount__, 1)

        self.source.int_prop = 10
        self.assertEqual(self.source.int_prop, 10)
        self.assertEqual(self.target.int_prop, 1)

        for i in range(10):
            binding.unbind()

    def test_reference_counts(self):
        self.assertEqual(self.source.__grefcount__, 1)
        self.assertEqual(self.target.__grefcount__, 1)

        # Binding ref count will be 2 do to the initial ref implicitly held by
        # the act of binding and the ref incurred by using __call__ to generate
        # a wrapper from the weak binding ref within python.
        binding = self.source.bind_property("int_prop", self.target, "int_prop")
        self.assertEqual(binding.__grefcount__, 2)

        # Creating a binding does not inc refs on source and target (they are weak
        # on the binding object itself)
        self.assertEqual(self.source.__grefcount__, 1)
        self.assertEqual(self.target.__grefcount__, 1)

        # Use GObject.get_property because the "props" accessor leaks.
        # Note property names are canonicalized.
        self.assertEqual(binding.get_property("source"), self.source)
        self.assertEqual(binding.get_property("source_property"), "int-prop")
        self.assertEqual(binding.get_property("target"), self.target)
        self.assertEqual(binding.get_property("target_property"), "int-prop")
        self.assertEqual(binding.get_property("flags"), GObject.BindingFlags.DEFAULT)

        # Delete reference to source or target and the binding will remove its own
        # "self reference".
        ref = self.source.weak_ref()
        del self.source
        gc.collect()
        self.assertEqual(ref(), None)
        self.assertEqual(binding.__grefcount__, 1)

        # Finally clear out the last ref held by the python wrapper
        ref = binding.weak_ref()
        del binding
        gc.collect()
        self.assertEqual(ref(), None)


class TestGValue(unittest.TestCase):
    def test_type_constant(self):
        self.assertEqual(GObject.TYPE_VALUE, GObject.Value.__gtype__)
        self.assertEqual(GObject.type_name(GObject.TYPE_VALUE), "GValue")

    def test_no_type(self):
        value = GObject.Value()
        self.assertEqual(value.g_type, GObject.TYPE_INVALID)
        self.assertRaises(TypeError, value.set_value, 23)
        self.assertEqual(value.get_value(), None)

    def test_int(self):
        value = GObject.Value(GObject.TYPE_UINT)
        self.assertEqual(value.g_type, GObject.TYPE_UINT)
        value.set_value(23)
        self.assertEqual(value.get_value(), 23)
        value.set_value(42.0)
        self.assertEqual(value.get_value(), 42)

    def test_string(self):
        value = GObject.Value(str, "foo_bar")
        self.assertEqual(value.g_type, GObject.TYPE_STRING)
        self.assertEqual(value.get_value(), "foo_bar")

    def test_float(self):
        # python float is G_TYPE_DOUBLE
        value = GObject.Value(float, 23.4)
        self.assertEqual(value.g_type, GObject.TYPE_DOUBLE)
        value.set_value(1e50)
        self.assertAlmostEqual(value.get_value(), 1e50)

        value = GObject.Value(GObject.TYPE_FLOAT, 23.4)
        self.assertEqual(value.g_type, GObject.TYPE_FLOAT)
        self.assertRaises(TypeError, value.set_value, "string")
        self.assertRaises(OverflowError, value.set_value, 1e50)

    def test_float_inf_nan(self):
        nan = float("nan")
        for type_ in [GObject.TYPE_FLOAT, GObject.TYPE_DOUBLE]:
            for x in [float("inf"), float("-inf"), nan]:
                value = GObject.Value(type_, x)
                # assertEqual() is False for (nan, nan)
                if x is nan:
                    self.assertEqual(str(value.get_value()), "nan")
                else:
                    self.assertEqual(value.get_value(), x)

    def test_enum(self):
        value = GObject.Value(Gio.FileType, Gio.FileType.DIRECTORY)
        self.assertEqual(value.get_value(), Gio.FileType.DIRECTORY)

    def test_flags(self):
        value = GObject.Value(Gio.FileCopyFlags, Gio.FileCopyFlags.OVERWRITE)
        self.assertEqual(value.get_value(), Gio.FileCopyFlags.OVERWRITE)

    def test_object(self):
        class TestObject(GObject.Object):
            pass

        obj = TestObject()
        value = GObject.Value(GObject.TYPE_OBJECT, obj)
        self.assertEqual(value.get_value(), obj)

    def test_value_array(self):
        value = GObject.Value(GObject.ValueArray)
        self.assertEqual(value.g_type, GObject.type_from_name("GValueArray"))
        value.set_value([32, "foo_bar", 0.3])
        self.assertEqual(value.get_value(), [32, "foo_bar", 0.3])

    def test_value_array_from_gvalue_list(self):
        value = GObject.Value(
            GObject.ValueArray,
            [
                GObject.Value(GObject.TYPE_UINT, 0xFFFFFFFF),
                GObject.Value(GObject.TYPE_STRING, "foo_bar"),
            ],
        )
        self.assertEqual(value.g_type, GObject.type_from_name("GValueArray"))
        self.assertEqual(value.get_value(), [0xFFFFFFFF, "foo_bar"])
        self.assertEqual(
            testhelper.value_array_get_nth_type(value, 0), GObject.TYPE_UINT
        )
        self.assertEqual(
            testhelper.value_array_get_nth_type(value, 1), GObject.TYPE_STRING
        )

    def test_value_array_append_gvalue(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)

            arr = GObject.ValueArray.new(0)
            arr.append(GObject.Value(GObject.TYPE_UINT, 0xFFFFFFFF))
            arr.append(GObject.Value(GObject.TYPE_STRING, "foo_bar"))
            self.assertEqual(arr.get_nth(0), 0xFFFFFFFF)
            self.assertEqual(arr.get_nth(1), "foo_bar")
            self.assertEqual(
                testhelper.value_array_get_nth_type(arr, 0), GObject.TYPE_UINT
            )
            self.assertEqual(
                testhelper.value_array_get_nth_type(arr, 1), GObject.TYPE_STRING
            )

    def test_gerror_boxing(self):
        error = GLib.Error("test message", domain="mydomain", code=42)
        value = GObject.Value(GLib.Error, error)
        self.assertEqual(value.g_type, GObject.type_from_name("GError"))

        unboxed = value.get_value()
        self.assertEqual(unboxed.message, error.message)
        self.assertEqual(unboxed.domain, error.domain)
        self.assertEqual(unboxed.code, error.code)

    def test_gerror_novalue(self):
        GLib.Error("test message", domain="mydomain", code=42)
        value = GObject.Value(GLib.Error)
        self.assertEqual(value.g_type, GObject.type_from_name("GError"))
        self.assertEqual(value.get_value(), None)


class TestDispose(unittest.TestCase):
    def test_dispose(self):
        class TestObject(GObject.Object):
            def __init__(self):
                super().__init__()
                self.dispose_invoked = False

            def do_dispose(self):
                self.dispose_invoked = True
                super().do_dispose()

        obj = TestObject()
        obj.run_dispose()

        assert obj.dispose_invoked

    def test_dispose_with_python_base_class(self):
        class TestBaseObject(GObject.Object):
            def __init__(self):
                super().__init__()
                self.dispose_invoked_base = False

            def do_dispose(self):
                self.dispose_invoked_base = True
                super().do_dispose()

        class TestObject(TestBaseObject):
            def __init__(self):
                super().__init__()
                self.dispose_invoked = False

            def do_dispose(self):
                self.dispose_invoked = True
                super().do_dispose()

        obj = TestObject()
        obj.run_dispose()

        assert obj.dispose_invoked_base
        assert obj.dispose_invoked

    @unittest.skipIf(platform.python_implementation() == "PyPy", "CPython only")
    def test_dispose_on_object_destruction(self):
        dispose_invoked = False

        class TestObject(GObject.Object):
            def do_dispose(self):
                nonlocal dispose_invoked
                dispose_invoked = True
                super().do_dispose()

        TestObject()

        assert dispose_invoked

    @unittest.skipIf(platform.python_implementation() == "PyPy", "CPython only")
    def test_dispose_on_floating_object(self):
        dispose_invoked = False

        class TestObject(GObject.Object):
            def do_dispose(self):
                nonlocal dispose_invoked
                dispose_invoked = True
                super().do_dispose()

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            TestObject()._force_floating()

        assert dispose_invoked

    def test_only_call_do_dispose_if_defined_on_class(self):
        dispose_invoked = False

        def do_dispose(self):
            nonlocal dispose_invoked
            dispose_invoked = True

        class TestObject(GObject.Object):
            def __init__(self):
                self.do_dispose = do_dispose
                super().__init__()

        TestObject()

        assert not dispose_invoked

    @unittest.skipIf(platform.python_implementation() == "PyPy", "CPython only")
    def test_access_members_objects_from_dispose(self):
        dispose_invoked = False

        class TestObject(GObject.Object):
            def __init__(self, a, b):
                super().__init__()
                self.a = a
                self.b = b

            def do_dispose(self):
                nonlocal dispose_invoked
                # access members: do_dispose should be called before they're removed
                self.a.val
                self.b.val
                dispose_invoked = True

        class TestMember:
            def __init__(self, val):
                self.val = val

        a = TestMember("a")
        b = TestMember("b")
        o = TestObject(a, b)
        a.o = o
        del a, b, o

        gc.collect()

        assert dispose_invoked


def test_list_properties():
    def find_param(props, name):
        for param in props:
            if param.name == name:
                return param
        return None

    list_props = GObject.list_properties

    props = list_props(Gio.Action)
    param = find_param(props, "enabled")
    assert param
    assert param.value_type == GObject.TYPE_BOOLEAN
    assert list_props("GAction") == list_props(Gio.Action)
    assert list_props(Gio.Action.__gtype__) == list_props(Gio.Action)

    props = list_props(Gio.SimpleAction)
    assert find_param(props, "enabled")

    def names(props):
        return [p.name for p in props]

    assert set(names(list_props(Gio.Action))) <= set(
        names(list_props(Gio.SimpleAction))
    )

    props = list_props(Gio.FileIcon)
    param = find_param(props, "file")
    assert param
    assert param.value_type == Gio.File.__gtype__

    assert list_props("GFileIcon") == list_props(Gio.FileIcon)
    assert list_props(Gio.FileIcon.__gtype__) == list_props(Gio.FileIcon)
    assert list_props(Gio.FileIcon(file=Gio.File.new_for_path("."))) == list_props(
        Gio.FileIcon
    )

    for obj in [Gio.ActionEntry, Gio.DBusError, 0, object()]:
        with pytest.raises(TypeError):
            list_props(obj)


def test_custom_class_update():
    # Example from Gaupol <https://github.com/otsaloma/gaupol>

    class CustomAction(Gio.SimpleAction):
        def __new__(cls):
            action = Gio.SimpleAction.new("custom-action")
            action.__class__ = cls
            return action

    action = CustomAction()
    assert type(action) is CustomAction

    group = Gio.SimpleActionGroup()
    group.insert(action)
    del action
    new_action = group.lookup("custom-action")

    assert type(new_action) is CustomAction



import pytest

from gi import PyGIDeprecationWarning
from gi.repository import GObject, GLib, GIMarshallingTests

from .helper import ignore_gi_deprecation_warnings


def test_stop_emission_deprec():
    class TestObject(GObject.GObject):
        int_prop = GObject.Property(default=0, type=int)

    obj = TestObject()

    def notify_callback(obj, *args):
        with pytest.warns(PyGIDeprecationWarning):
            obj.stop_emission("notify::int-prop")

        with pytest.warns(PyGIDeprecationWarning):
            obj.emit_stop_by_name("notify::int-prop")

        obj.stop_emission_by_name("notify::int-prop")

    obj.connect("notify::int-prop", notify_callback)
    obj.notify("int-prop")


def test_signal_parse_name():
    obj = GObject.GObject()
    assert GObject.signal_parse_name("notify", obj, True) == (1, 0)

    with pytest.raises(ValueError):
        GObject.signal_parse_name("foobar", obj, True)


def test_signal_query():
    obj = GObject.GObject()
    res = GObject.signal_query("notify", obj)
    assert res.signal_name == "notify"
    assert res.itype == obj.__gtype__

    res = GObject.signal_query("foobar", obj)
    assert res is None


def test_value_repr():
    v = GObject.Value()
    assert repr(v) == "<Value (invalid) None>"

    v = GObject.Value(int, 0)
    assert repr(v) == "<Value (gint) 0>"


def test_value_no_init():
    v = GObject.Value()
    with pytest.raises(TypeError):
        v.set_value(0)
    v.init(GObject.TYPE_LONG)
    assert v.get_value() == 0
    v.set_value(0)


def test_value_invalid_type():
    v = GObject.Value()
    assert v.g_type == GObject.TYPE_INVALID
    assert isinstance(GObject.TYPE_INVALID, GObject.GType)
    with pytest.raises(ValueError, match="Invalid GType"):
        v.init(GObject.TYPE_INVALID)

    with pytest.raises(TypeError, match=r"GObject.Value needs to be initialized first"):
        v.set_value(None)

    assert v.get_value() is None


def test_value_long():
    v = GObject.Value(GObject.TYPE_LONG)
    assert v.get_value() == 0
    v.set_value(0)
    assert v.get_value() == 0

    v.set_value(GLib.MAXLONG)
    assert v.get_value() == GLib.MAXLONG

    v.set_value(GLib.MINLONG)
    assert v.get_value() == GLib.MINLONG

    with pytest.raises(OverflowError):
        v.set_value(GLib.MAXLONG + 1)

    with pytest.raises(OverflowError):
        v.set_value(GLib.MINLONG - 1)


def test_value_ulong():
    v = GObject.Value(GObject.TYPE_ULONG)
    assert v.get_value() == 0
    v.set_value(0)
    assert v.get_value() == 0

    v.set_value(GLib.MAXULONG)
    assert v.get_value() == GLib.MAXULONG

    with pytest.raises(OverflowError):
        v.set_value(GLib.MAXULONG + 1)

    with pytest.raises(OverflowError):
        v.set_value(-1)

    with pytest.raises(TypeError):
        v.set_value(object())

    with pytest.raises(TypeError):
        v.set_value(None)


def test_value_float():
    v = GObject.Value(GObject.TYPE_FLOAT)

    for getter, setter in [(v.get_value, v.set_value), (v.get_float, v.set_float)]:
        assert getter() == 0.0
        setter(0)
        assert getter() == 0

        setter(GLib.MAXFLOAT)
        assert getter() == GLib.MAXFLOAT

        setter(GLib.MINFLOAT)
        assert getter() == GLib.MINFLOAT

        setter(-GLib.MAXFLOAT)
        assert getter() == -GLib.MAXFLOAT

        with pytest.raises(OverflowError):
            setter(GLib.MAXFLOAT * 2)

        with pytest.raises(OverflowError):
            setter(-GLib.MAXFLOAT * 2)

        with pytest.raises(TypeError):
            setter(object())

        with pytest.raises(TypeError):
            setter(None)

        with pytest.raises(TypeError):
            setter(1j)

        v.reset()


def test_value_double():
    v = GObject.Value(GObject.TYPE_DOUBLE)
    assert v.get_value() == 0.0
    v.set_value(0)
    assert v.get_value() == 0

    v.set_value(GLib.MAXDOUBLE)
    assert v.get_value() == GLib.MAXDOUBLE

    v.set_value(GLib.MINDOUBLE)
    assert v.get_value() == GLib.MINDOUBLE

    v.set_value(-GLib.MAXDOUBLE)
    assert v.get_value() == -GLib.MAXDOUBLE

    with pytest.raises(TypeError):
        v.set_value(object())

    with pytest.raises(TypeError):
        v.set_value(None)

    with pytest.raises(TypeError):
        v.set_value(1j)


def test_value_uint64():
    v = GObject.Value(GObject.TYPE_UINT64)
    assert v.get_value() == 0
    v.set_value(0)
    assert v.get_value() == 0

    v.set_value(GLib.MAXUINT64)
    assert v.get_value() == GLib.MAXUINT64

    with pytest.raises(OverflowError):
        v.set_value(GLib.MAXUINT64 + 1)

    with pytest.raises(OverflowError):
        v.set_value(-1)


def test_value_int64():
    v = GObject.Value(GObject.TYPE_INT64)
    assert v.get_value() == 0
    v.set_value(0)
    assert v.get_value() == 0

    v.set_value(GLib.MAXINT64)
    assert v.get_value() == GLib.MAXINT64
    v.set_value(GLib.MININT64)
    assert v.get_value() == GLib.MININT64

    with pytest.raises(OverflowError):
        v.set_value(GLib.MAXINT64 + 1)

    with pytest.raises(OverflowError):
        v.set_value(GLib.MININT64 - 1)

    with pytest.raises(TypeError):
        v.set_value(object())

    with pytest.raises(TypeError):
        v.set_value(None)


def test_value_pointer():
    v = GObject.Value(GObject.TYPE_POINTER)
    assert v.get_value() == 0
    v.set_value(42)
    assert v.get_value() == 42
    v.set_value(0)
    assert v.get_value() == 0


def test_value_unichar():
    assert GObject.TYPE_UNICHAR == GObject.TYPE_UINT

    v = GObject.Value(GObject.TYPE_UNICHAR)
    assert v.get_value() == 0
    v.set_value(42)
    assert v.get_value() == 42

    v.set_value(GLib.MAXUINT)
    assert v.get_value() == GLib.MAXUINT


def test_value_gtype():
    class TestObject(GObject.GObject):
        pass

    v = GObject.Value(GObject.TYPE_GTYPE)
    assert v.get_value() == GObject.TYPE_INVALID
    v.set_value(TestObject.__gtype__)
    assert v.get_value() == TestObject.__gtype__
    v.set_value(TestObject)
    assert v.get_value() == TestObject.__gtype__

    with pytest.raises(TypeError):
        v.set_value(None)


def test_value_variant():
    v = GObject.Value(GObject.TYPE_VARIANT)
    assert v.get_value() is None
    variant = GLib.Variant("i", 42)
    v.set_value(variant)
    assert v.get_value() == variant

    v.set_value(None)
    assert v.get_value() is None

    with pytest.raises(TypeError):
        v.set_value(object())


def test_value_param():
    # FIXME: set_value and get_value trigger a critical
    # GObject.Value(GObject.TYPE_PARAM)
    pass


def test_value_string():
    v = GObject.Value(GObject.TYPE_STRING)
    for getter, setter in [(v.get_value, v.set_value), (v.get_string, v.set_string)]:
        assert getter() is None

        with pytest.raises(TypeError):
            setter(b"bar")

        setter("quux")
        assert getter() == "quux"
        assert isinstance(getter(), str)

        setter(None)
        assert getter() is None

        v.reset()


def test_value_pyobject():
    class Foo:
        pass

    v = GObject.Value(GObject.TYPE_PYOBJECT)
    assert v.get_value() is None
    for obj in [Foo(), None, 42, "foo"]:
        v.set_value(obj)
        assert v.get_value() == obj


@ignore_gi_deprecation_warnings
def test_value_char():
    v = GObject.Value(GObject.TYPE_CHAR)
    assert v.get_value() == 0
    v.set_value(42)
    assert v.get_value() == 42
    v.set_value(-1)
    assert v.get_value() == -1
    v.set_value(b"a")
    assert v.get_value() == 97
    v.set_value(b"\x00")
    assert v.get_value() == 0

    v.set_value("a")
    assert v.get_value() == 97

    with pytest.raises(OverflowError):
        v.set_value(128)


def test_value_uchar():
    v = GObject.Value(GObject.TYPE_UCHAR)
    assert v.get_value() == 0
    v.set_value(200)
    assert v.get_value() == 200
    v.set_value(b"a")
    assert v.get_value() == 97
    v.set_value(b"\x00")
    assert v.get_value() == 0

    v.set_value("a")
    assert v.get_value() == 97

    with pytest.raises(OverflowError):
        v.set_value(256)


def test_value_set_boxed_deprecate_non_boxed():
    v = GObject.Value(GObject.TYPE_POINTER)
    with pytest.warns(PyGIDeprecationWarning):
        v.get_boxed()
    with pytest.warns(PyGIDeprecationWarning):
        v.set_boxed(None)


def test_value_boolean():
    v = GObject.Value(GObject.TYPE_BOOLEAN)
    for getter, setter in [(v.get_value, v.set_value), (v.get_boolean, v.set_boolean)]:
        assert getter() is False
        assert isinstance(getter(), bool)

        setter(42)
        assert getter() is True
        setter(-1)
        assert getter() is True
        setter(0)
        assert getter() is False

        setter([])
        assert getter() is False
        setter(["foo"])
        assert getter() is True

        setter(None)
        assert getter() is False
        v.reset()


def test_value_enum():
    t = GIMarshallingTests.GEnum
    v = GObject.Value(t)

    for getter, setter in [(v.get_value, v.set_value), (v.get_enum, v.set_enum)]:
        assert v.g_type == t.__gtype__
        assert getter() == 0

        setter(t.VALUE1)
        assert getter() == t.VALUE1
        # FIXME: we should try to return an enum type
        assert type(getter()) is int

        setter(2424242)
        assert getter() == 2424242

        setter(-1)
        assert getter() == -1

        with pytest.raises(TypeError):
            setter(object())

        with pytest.raises(TypeError):
            setter(None)

        v.reset()


def test_value_object():
    v = GObject.Value(GIMarshallingTests.Object)
    assert v.g_type.is_a(GObject.TYPE_OBJECT)

    for getter, setter in [(v.get_value, v.set_value), (v.get_object, v.set_object)]:
        assert getter() is None

        setter(None)
        assert getter() is None

        obj = GIMarshallingTests.Object()
        setter(obj)
        assert getter() is obj

        with pytest.raises(TypeError):
            setter(object())

        v.reset()
